"""Ground state file parsers.

All functions in this module could benefit from refactoring.
"""
import xml.etree.ElementTree as ET
import re
import numpy as np

from excitingtools.parser_utils.erroneous_file_error import ErroneousFileError
from excitingtools.parser_utils.parser_decorators import xml_root


def parse_info_out(name: str) -> dict:
    """
    Parser exciting INFO.OUT into a dictionary.
    In:
        name     string     path of the file to parse
    Out:
        info     dict       contains the content of the file to parse
    """
    file = open(name)
    lines = file.readlines()
    file.close()

    nscl = []
    nini = []

    # Get line numbers for SCF iteration blocks
    for i, line in enumerate(lines):
        # stores the number of the first and last line of every iteration into a list
        if ('SCF iteration number' in line) or ('Hybrids iteration number'in line) \
           or ('Reached self-consistent loops maximum' in line):
            nscl.append(i)
        if ('Convergence criteria checked for the last 2 iterations'
            in line) or ('Self-consistent loop stopped' in line):
            nscl.append(i)
        # stores the number of the first and last line of the initialization into a list
        if 'Starting initialization' in line:
            nini.append(i + 2)
        if 'Ending initialization' in line:
            nini.append(i - 2)

    calculation_failed = True
    for line in reversed(lines):
        if ("EXCITING" in line) and ("stopped" in line):
            calculation_failed = False
            break

    if calculation_failed:
        raise ErroneousFileError()

    INFO = {}
    INFO['initialization'] = {}
    ini = []
    inits = {}
    k = 0
    speci = 0  # variable to detect different species in INFO.OUT

    # loops through all lines of the initialization
    for i in range(nini[0], nini[1]):
        # stores the lines, which have the format "variable : value" into a list
        if (':' in lines[i]):
            lines[i] = lines[i].split(':')
            ini.append(lines[i])
            if ini[k][0][1] != ' ' and speci != 0:  # if indentation stops, species part is ended
                speci = 0

            ini[k][0] = ini[k][0].strip()
            ini[k][1] = ini[k][1].strip()
            if ('Lattice vectors'
                in ini[k][0]) or ('Reciprocal lattice vectors'
                                  in ini[k][0]):
                ini[k][1] = []
                lines[i + 1] = (lines[i + 1].split())
                lines[i + 2] = (lines[i + 2].split())
                lines[i + 3] = (lines[i + 3].split())
                for j in range(3):
                    ini[k][1].append(lines[i + 1][j])
                    ini[k][1].append(lines[i + 2][j])
                    ini[k][1].append(lines[i + 3][j])
                if ' ' in ini[k][1]:
                    ini[k][1] = ini[k][1].split()

            # initialize species subdict if key Species is found:
            if ini[k][0] == 'Species':
                speci = ini[k][1][0]
                speci_name = ini[k][1][3:-1]
                inits.update({'Species ' + speci: {'Species symbol': speci_name}})

            # stores variable-value pairs in a dictionary, in species subdict if necessary
            if speci != 0:
                if ini[k][0][:16] == 'atomic positions':
                    split_key = ini[k][0].split()
                    unit = split_key[2][1:-1]
                    inits['Species ' + speci].update({'Atomic positions': {}})
                else:
                    try:
                        key_name = 'Atom ' + str(int(ini[k][0]))
                        inits['Species ' + speci]['Atomic positions'].update({key_name: ini[k][1]})
                    except ValueError:
                        inits['Species ' + speci].update({ini[k][0]: ini[k][1]})
            else:
                inits.update({ini[k][0]: ini[k][1]})
            k = k + 1
        # type of mixing is stored in the dictionary too
        if 'mixing' in lines[i]:
            lines[i] = lines[i].strip()
            inits.update({'mixing': lines[i]})
    inits.update({'units': {'positions': unit}})

    INFO['initialization'] = inits

    scl1 = []
    INFO['scl'] = {}

    # loops through all scl's
    for j in range(len(nscl) - 1):
        scls = {}
        scl = []
        k = 0
        # loops through all lines of the scl
        for i in range(nscl[j], nscl[j + 1]):
            # stores the lines, which have the format "variable : value" into a list
            if (':' in lines[i]) and ('+'
                                      not in lines[i]) and ('(target)'
                                                            not in lines[i]):
                lines[i] = lines[i].split(':')
                scl.append(lines[i])
                scl[k][0] = scl[k][0].strip()
                scl[k][1] = scl[k][1].strip()
                if ' ' in scl[k][1]:
                    scl[k][1] = scl[k][1].split()
                # stores variable-value pairs in a dictionary
                scls.update({scl[k][0]: scl[k][1]})
                k = k + 1
        INFO['scl'][str(j + 1)] = scls

    return INFO


def parse_info_xml(name) -> dict:
    """
    Parser exciting info.xml into a python dictionary.
    In:
        name     string     path of the file to parse
    Out:
        info     dict       contains the content of the file to parse
    """
    try:
        root = ET.parse(name)
    except AttributeError:
        raise ErroneousFileError

    info = root.find('groundstate').attrib

    excitingRun = []
    i = 0
    for node in root.find('groundstate').find('scl').iter('iter'):
        excitingRun.append(node.attrib)
        excitingRun[i]['energies'] = node.find('energies').attrib
        excitingRun[i]['charges'] = node.find('charges').attrib
        atom_nr = 0
        atomic_charge = []
        species = []
        for atoms in node.find('charges').iter('atom'):
            if atom_nr == 0: species_old = atoms.get('species')
            atom_nr = atom_nr + 1
            if atoms.get('species') == species_old:
                species.append({'muffin-tin': atoms.get('muffin-tin')})
            else:
                species_old = atoms.get('species')
                atomic_charge.append(species)
                species = [{'muffin-tin': atoms.get('muffin-tin')}]
                atomic_charge.append(species)
                excitingRun[i]['charges']['atomic'] = atomic_charge
                excitingRun[i]['timing'] = node.find('timing').attrib
        if node.find('moments') is not None:
            moments = {}
            moments['momtot'] = node.find('moments').find('momtot').attrib
            moments['interstitial'] = node.find('moments').find(
                'momtot').attrib
            moments['mommttot'] = node.find('moments').find(
                'interstitial').attrib
            excitingRun[i]['moments'] = moments
            atom_nr = 0
            atomic_moment = []
            species = []
            for atoms in node.find('moments').iter('atom'):
                if atom_nr == 0: species_old = atoms.get('species')
                atom_nr = atom_nr + 1
                if atoms.get('species') == species_old:
                    species.append(atoms.find('mommt').attrib)
                else:
                    species_old = atoms.get('species')
                    atomic_moment.append(species)
                    species = [atoms.find('mommt').attrib]
                    atomic_moment.append(species)
                    excitingRun[i]['moments']['atomic'] = atomic_moment
        i = i + 1
    info['scl'] = {}
    for item in excitingRun:  # converts list of scl-iterations into a dictionary
        name = item['iteration']
        info['scl'][name] = item

    return info


def parse_atoms(name) -> dict:
    """                                                                                                  
    Parser exciting atoms.xml into a python dictionary.                                                   
    In:                                                                                                  
        name     string     path of the file to parse                                                    
    Out:                                                                                                 
        info     dict       contains the content of the file to parse                                    
    """

    root = ET.parse(name)
    atoms = {}
    atoms['Hamiltonian'] = root.find('Hamiltonian').attrib
    atom = []
    i = 0
    for node in root.findall('atom'):
        atom.append(node.attrib)

        spectrum = []
        states = node.find('spectrum')
        for state in states.findall('state'):
            spectrum.append(state.attrib)

        atom[i]['NumericalSetup'] = node.find('NumericalSetup').attrib
        atom[i]['spectrum'] = {}
        j = 0
        for item in spectrum:  # converts list of states into a dictionary
            name = str(j)
            atom[i]['spectrum'][name] = item
            j = j + 1
        i = i + 1
    atoms['atom'] = {}
    for item in atom:  # converts list of atoms into a dictionary
        name = item['chemicalSymbol']
        atoms['atom'][name] = item

    return atoms


@xml_root
def parse_eigval(root) -> dict:
    """ Parse eigenvalues from eigval.xml file.

    :param root: XML file name, XML string or ElementTree.Element as input.
    :return: dict output: Parsed data.
    """
    eigval = root.attrib

    kpts = []
    for node in root.findall('kpt'):
        kpt = node.attrib
        state = []
        for subnode in node:
            state.append(subnode.attrib)
            kpt['state'] = {}  # converts list of states into a dictionary
        for item in state:
            name = item['ist']
            kpt['state'][name] = item
            kpts.append(kpt)
            eigval['kpt'] = {}
    for item in kpts:  # converts list of kpts into a dictionary
        name = item['ik']
        eigval['kpt'][name] = item

    return eigval


def parse_evalcore(name) -> dict:
    """                                                                                                  
    Parser exciting evalcore.xml into a python dictionary.                                                   
    In:                                                                                                  
        name     string     path of the file to parse                                                    
    Out:                                                                                                 
        info     dict       contains the content of the file to parse                                    
    """

    root = ET.parse(name).getroot()
    evalcore = root.attrib

    speciess = []
    for node in root.findall('species'):
        species = node.attrib
        atoms = []
        for subnode in node:
            atom = subnode.attrib
            states = []
            for subnode1 in subnode:
                state = subnode1.attrib
                states.append(state)
            atom['state'] = {}
            for item in states:  # converts list of states into a dictionary
                name = item['ist']
                atom['state'][name] = item
            atoms.append(atom)
            species['atom'] = {}
            for item in atoms:  # converts list of atoms into a dictionary
                name = item['ia']
                species['atom'][name] = item
        speciess.append(species)
    evalcore['species'] = {}
    for item in speciess:  # converts list of species into a dictionary
        name = item['chemicalSymbol']
        evalcore['species'][name] = item

    return evalcore


def parse_geometry(name) -> dict:
    """                                                                                                  
    Parser exciting geometry.xml into a python dictionary.                                                   
    In:                                                                                                  
        name     string     path of the file to parse                                                    
    Out:                                                                                                 
        info     dict       contains the content of the file to parse                                    
    """

    root = ET.parse(name).getroot()
    structure = root.find('structure').attrib
    crystal = root.find('structure').find('crystal').attrib
    geometry = {'structure': structure}
    structure['crystal'] = crystal
    speciess = []
    for node in root.find('structure').findall('species'):
        species = node.attrib
        atoms = []
        for subnode in node:
            atom = subnode.attrib
            atoms.append(atom)
            species['atom'] = {}
        i = 1
        for item in atoms:
            name = str(i)
            species['atom'][name] = item['coord'].split()
            i = i + 1
        speciess.append(species)
    structure['species'] = {}
    j = 1
    for item in speciess:
        name = str(j)
        structure['species'][name] = item
        j = j + 1

    basevects = []
    for node in root.find('structure').find('crystal').findall('basevect'):
        basevect = node.text
        basevects.append(basevect)
        structure['crystal']['basevect'] = {}
        k = 1
    for item in basevects:
        name = str(k)
        structure['crystal']['basevect'][name] = item
        k = k + 1
    return geometry


def parse_linengy(name: str) -> dict:
    """
    Parser for: LINENGY.OUT

    :param name: path of the file to parse
    :returns: dictionary containing parsed file
    """

    linengy = {}

    with open(file=name, mode='r') as fid:
        lines = fid.readlines()

    apw_line = []
    lo_line = []

    for i, line in enumerate(lines):
        if 'APW functions' in line:
            apw_line.append(i)
        if 'local-orbital functions' in line:
            lo_line.append(i)

    apw_line.append(len(lines))

    for i in range(0, len(lo_line)):
        linengy[str(i)] = {}
        apw = []
        lo = []

        for j in range(apw_line[i], lo_line[i]):
            if "l =" in lines[j]:
                apw.append(lines[j].split(':')[1].strip())
        linengy[str(i)]['apw'] = apw

        for j in range(lo_line[i], apw_line[i+1]):
            if "l =" in lines[j]:
                lo.append(lines[j].split(':')[1].strip())
        linengy[str(i)]['lo'] = lo

    return linengy


def parse_lo_recommendation(name: str) -> dict:
    """
    Parser for: LO_RECOMMENDATION.OUT

    :param name: path of the file to parse
    :returns: dictionary containing parsed file                                                                                                  
    """
    with open(file=name, mode='r') as fid:
        lines = fid.readlines()

    n_species = int(lines[2].split(':')[1])
    n_l_channels = int(lines[3].split(':')[1])
    n_nodes = int(lines[4].split(':')[1])

    lo_recommendation = {'n_species': n_species, 'n_l_channels': n_l_channels, 'n_nodes': n_nodes}

    blocks = lines[6 :]
    block_size = n_nodes + 3
    n_blocks = n_l_channels * n_species

    for iblock in range(0, n_blocks):
        offset = iblock * block_size
        matches = re.findall(r':\s(.*?)(?:,|$)', blocks[offset+0])
        species, l = matches[0], int(matches[1])
        # Package (node, n, energy) as one likes                                                                                        
        if not species in lo_recommendation:
            lo_recommendation.update({species: {}})
        lo_recommendation[species].update({l: np.loadtxt(blocks[offset+2:offset+n_nodes+2]).tolist()})

    return lo_recommendation




""" Utilities for interaction with the ASE library. """

from ase import Atoms

from excitingtools import ExcitingStructure
from excitingtools.constants.units import bohr_to_angstrom


def exciting_structure_to_ase(structure: ExcitingStructure) -> Atoms:
    """ Function to extract the physical structure from an exciting structure object
    and transforms it into an ase.atoms.Atoms object.

    :param structure: input exciting structure object
    :returns: ASE Atoms object
    """
    lattice = structure.get_lattice(convert_to_angstrom=True)
    if structure.structure_attributes.get("cartesian"):
        positions = structure.positions * bohr_to_angstrom
        return Atoms(symbols=structure.species, positions=positions, cell=lattice, pbc=True)
    else:
        return Atoms(symbols=structure.species, scaled_positions=structure.positions, cell=lattice, pbc=True)



""" Simple text parsers.
"""
from typing import Union


def match_current_return_line_n(file_string: str, match: str, n_line=1) -> Union[str, None]:
    """
    Match a string on the ith line and return a substring from the i+n_line line.

    :param str file_string: Input string
    :param str match: String to match
    :param int n_line: The index of the line to return, following the matched line

    :return Union[str, None] matched line string, or None
    """
    file = file_string.split('\n')
    for i, line in enumerate(file):
        if match in line:
            return file[i + n_line]
    return None


def match_current_extract_from_line_n(input_string: str,
                                      keys_extractions: dict,
                                      n_line=1) -> dict:
    """
    Given an input_string, match a substring (defined by the key of keys_extractions),
    return the a substring from the i+n_line line below the match, and extract a value
    from that line using the value of the key, defined by keys_extractions[key]

    :param str file_string: Input string
    :param dict keys_extractions: keys = strings to match
                                  values = methods of data extraction from the strings
    :param int n_line: The index of the line to return, following the matched line

    :return dict data: Dictionary of extracted data from the nth line below a matched string.
    """
    data = {}
    for key, extract_data in keys_extractions.items():
        match = match_current_return_line_n(input_string, key, n_line=n_line)
        if match is not None:
            data[key] = extract_data(match)
    return data



""" VXCNN.DAT Parser.
"""
import numpy as np
import re
from typing import Union
import pathlib

from excitingtools.dataclasses.data_structs import NumberOfStates
from excitingtools.exciting_dict_parsers.gw_eigenvalues_parser import n_states_from_file


def vkl_from_vxc(file_string: str) -> dict:
    """Extract vkl (k-points in fractional coordinates) from VXCNN.DAT.

    Each k-point header is defined like:
     ik=   1    vkl=  0.0000  0.0000  0.0000

    k-indices are not extracted as they are always contiguous,
    with indexing starting at 1.

    :param str file_string: File string.
    :return dict vkl: k-points in fractional coordinates.
    """
    raw_data: list = re.findall(r'\s*ik= .*$', file_string, flags=re.MULTILINE)
    vkl = {}
    for ik, line in enumerate(raw_data):
        vkl[ik + 1] = [float(k) for k in line.split()[-3:]]
    return vkl


def parse_vxnn_vectors(full_file_name: Union[str, pathlib.Path], vkl: dict, n_states: int) -> dict:
    """Parse VXC diagonal matrix elements from VXCNN.DAT.

    The routine exploits the repeating file structure:

     ik=   1    vkl=  0.0000  0.0000  0.0000
        1       -2.908349       -0.000000
        2       -2.864103        0.000000
        3       -2.864103       -0.000000
        .
        n_states

     ik=   2    vkl=  0.0000  0.0000  0.5000
        1       -2.908349        0.000000
        2       -2.864100       -0.000000
        3       -2.864100        0.000000
        .
        n_states

    :param str full_file_name: Path + file name
    :param dict vkl: Dictionary of vkl (k-points?)
    :param int n_states: Total number of occupied plus empty states
     Note, this is constant per q-point.

    :return dict data: Parsed VXC diagonal matrix elements, per k-point.
    """
    # File formatting
    header_size = 1
    blank_line = 1

    data = {}
    skip_lines = header_size

    # Must iterate lowest to highest, else data won't match k-points
    for ik in range(1, len(vkl) + 1):
        vxc_vector = np.loadtxt(full_file_name,
                                skiprows=skip_lines,
                                max_rows=n_states)
        # Ignore first column (state index)
        data[ik] = vxc_vector[:, 1:]
        skip_lines += n_states + (header_size + blank_line)

    return data


def n_states_from_vxcnn(file_string: str) -> NumberOfStates:
    return n_states_from_file(file_string, n_header=1)


def parse_vxcnn(full_file_name: Union[str, pathlib.Path]) -> dict:
    """Parser for VXCNN.DAT.

    :param str full_file_name: Path + file name
    :return dict data: Parsed k-points (labelled as vkl) and the diagonal elements of
     Vxc, per k-point.
    """
    try:
        with open(full_file_name) as f:
            file_string = f.read()
    except FileNotFoundError:
        raise FileNotFoundError(f"{full_file_name} does not exist")

    states = n_states_from_vxcnn(file_string)
    vkl = vkl_from_vxc(file_string)
    v_xc = parse_vxnn_vectors(full_file_name, vkl, states.n_states)
    assert len(vkl) == len(v_xc), (
        "Should be a vector of Vxc_NN for each k-point")

    # Repackage Vxc vectors with their respective k-points
    data = {}
    for ik in range(1, len(vkl) + 1):
        data[ik] = {'vkl': vkl[ik], 'v_xc_nn': v_xc[ik]}

    return data



from excitingtools.dataclasses.band_structure import BandData
import pytest
import numpy as np


@pytest.fixture
def band_data():
    """ Initialise an instance of BandData, and check attributes are set correctly.
    Reference data taken from 'bandstructure.xml' and 'bandstructure.dat' files for silver, containing only two bands
    and only 6 k-sampling points per band.
    """

    ref_bands = np.array([[-3.37071333, 0.59519479],
                          [-3.37071074, 0.59537998],
                          [-3.37070519, 0.59575556],
                          [-3.3706986, 0.59616303],
                          [-3.3706822, 0.59664939],
                          [-3.37066123, 0.59639211]])

    ref_k_points = np.array([[1.000000, 0.000000, 0.000000],
                             [0.988281, 0.011719, 0.000000],
                             [0.976562, 0.023438, 0.000000],
                             [0.964844, 0.035156, 0.000000],
                             [0.953125, 0.046875, 0.000000],
                             [0.941406, 0.058594, 0.000000]])

    ref_e_fermi = 0.0

    ref_flattened_k_points = np.array([0., 0.02697635, 0.05395270, 0.08092905, 0.10790540, 0.13488176])

    ref_vertices = [{'distance': 0.0, 'label': 'Gamma', 'coord': [0.0, 0.0, 0.0]},
                    {'distance': 0.8632432750, 'label': 'K', 'coord': [0.625, 0.375, 0.0]},
                    {'distance': 1.150991033, 'label': 'X', 'coord': [0.5, 0.5, 0.0]},
                    {'distance': 1.964864598, 'label': 'G', 'coord': [0.0, 0.0, 0.0]},
                    {'distance': 2.669699781, 'label': 'L', 'coord': [0.5, 0.0, 0.0]}]

    band_data = BandData(ref_bands, ref_k_points, ref_e_fermi, ref_flattened_k_points, ref_vertices)

    assert band_data.n_k_points == band_data.bands.shape[0], (
        "First dim of bands array equals the number of k-sampling points in the band structure")
    assert band_data.n_k_points == 6, "sampling points per band"
    assert band_data.n_bands == 2, "band_structure_xml contains two bands"
    assert np.allclose(band_data.k_points, ref_k_points, atol=1.e-8)
    assert np.allclose(band_data.bands, ref_bands, atol=1.e-8)
    assert band_data.vertices == ref_vertices

    return band_data


def test_xticks_and_labels(band_data):
    flattened_high_sym_points_ref = [0.000000000, 0.86324327, 1.15099103, 1.9648646, 2.66969978]
    unicode_gamma = '\u0393'
    labels_ref = [unicode_gamma, "K", "X", unicode_gamma, "L"]

    flattened_high_sym_points, labels = band_data.band_path()

    assert np.allclose(flattened_high_sym_points, flattened_high_sym_points_ref)
    assert labels == labels_ref


def test_get_band_edges(band_data):
    i_vbm, i_cbm = band_data.get_band_edges()
    assert i_vbm == 0
    assert i_cbm == 1

    band_data.e_fermi = 1.0
    with pytest.raises(ValueError, match="Fermi level 1.0 larger than highest band energy 0.59664939"):
        band_data.get_band_edges(), "ValueError is returned in case of erroneously large Fermi level"


def test_get_valence_band_maximum(band_data):
    vbm = band_data.get_valence_band_maximum()
    assert vbm == -3.37066123


def test_get_conduction_band_minimum(band_data):
    cbm = band_data.get_conduction_band_minimum()
    assert cbm == 0.59519479


def test_get_band_gap(band_data):
    k_valence = [1.000000, 0.000000, 0.000000]
    k_conduction = [0.941406, 0.058594, 0.000000]
    band_gap = band_data.get_band_gap(k_valence, k_conduction)
    assert band_gap == 3.96710544


def test_get_fundamental_band_gap(band_data):
    fundamental_gap = band_data.get_fundamental_band_gap()
    assert fundamental_gap == 3.96585602

    band_data.bands[0, 0] = 1
    fundamental_gap_zero = band_data.get_fundamental_band_gap()
    assert fundamental_gap_zero == 0.0



""" GW_INFO.OUT Parser, and all sub-block parses that comprise it.

NOTE. This could be trivialised to one line if GW_INFO.OUT was refactored
to write to YAML.
"""
import numpy as np
from typing import List, Union
import sys
import copy
import re

from excitingtools.parser_utils.parser_decorators import accept_file_name
from excitingtools.parser_utils.regex_parser import parse_value_regex, parse_values_regex
from excitingtools.parser_utils.simple_parser import match_current_return_line_n, match_current_extract_from_line_n
from excitingtools.utils.utils import can_be_float

# Info file name for GW
_file_name = "GW_INFO.OUT"


def parse_correlation_self_energy_params(file_string: str) -> dict:
    """Parse correlation self-energy parameters.

    Match the key on one line and return the value from the second, according to some
    specified extraction behaviour given by `keys_extractions`.

    Also extract the analytic continuation and singularity schemes, plus their references.

    :param str file_string: Input string
    :return dict data: Matched data
    """
    keys_extractions = {'Solution of the QP equation': lambda x: int(x.split()[0]),
                        'Energy alignment': lambda x: int(x.split()[0]),
                        'Analytic continuation method': lambda x: x.strip(),
                        'Scheme to treat singularities': lambda x: x.strip()
                        }

    data = match_current_extract_from_line_n(file_string, keys_extractions)

    # Also extract citations
    def ref_after_citation(x: str) -> str:
        """
        Expected format is
        Citation: Paper reference
        """
        try:
            i = x.index(":")
            return x[i + 1:].strip()
        except ValueError:
            return 'Not found'

    # Extraction of citations after
    for key in ['Analytic continuation method', 'Scheme to treat singularities']:
        citation_extraction = {key: ref_after_citation}
        matched_dictionary = match_current_extract_from_line_n(file_string, citation_extraction, n_line=2)
        data[key + ' citation'] = matched_dictionary[key]

    return data


def parse_mixed_product_params(file_string: str) -> dict:
    """Parse mixed product basis parameters.

    :param str file_string: Input string
    :return dict data: Matched data
    """
    search_keys = [
        'Angular momentum cutoff:', 'Linear dependence tolerance factor:',
        'Plane wave cutoff \\(in units of Gkmax\\):'
    ]

    data = parse_values_regex(file_string, search_keys)

    # Rename keys to give better context
    modified_data = {
        'MT Angular momentum cutoff':
        data['Angular momentum cutoff'],
        'MT Linear dependence tolerance factor':
        data['Linear dependence tolerance factor'],
        'Plane wave cutoff (in units of Gkmax)':
        data['Plane wave cutoff (in units of Gkmax)']
    }

    return modified_data


def parse_bare_coulomb_potential_params(file_string: str) -> dict:
    """Extract bare Coulomb parameter data of the form:

    Bare Coulomb potential parameters:
        Plane wave cutoff (in units of Gkmax*input%gw%MixBasis%gmb): 2.00000000000000
        Error tolerance for structure constants:   1.000000000000000E-016
        Tolerance factor to reduce the MB size based on
        the eigenvectors of the bare Coulomb potential:   0.100000000000000

    and return a dictionary of the form:

    {'Plane wave cutoff (in units of Gkmax*gmb)': 2.0,
     'Error tolerance for structure constants': 1e-16,
      'MB tolerance factor': 0.1
    }

    :param str file_string: Input string
    :return dict data: Matched data
    """

    pw_cutoff_matching_str = 'Plane wave cutoff \\(in units of Gkmax\\*input%gw%MixBasis%gmb\\):'
    pw_cutoff = list(parse_value_regex(file_string, pw_cutoff_matching_str).values())
    assert len(pw_cutoff) == 1, "Matched plane wave cutoff is ambiguous - more than one match"

    # Defined with less-verbose key
    data = {'Plane wave cutoff (in units of Gkmax*gmb)': float(pw_cutoff[0])}

    data2 = parse_value_regex(file_string, 'Error tolerance for structure constants:')

    data3 = parse_value_regex(file_string, 'the eigenvectors of the bare Coulomb potential:')

    # Defined with less-verbose key (see docs above for full key, over 2 lines)
    modified_data3 = {'MB tolerance factor': data3.pop('the eigenvectors of the bare Coulomb potential')}

    return {**data, **data2, **modified_data3}


def parse_mixed_product_wf_info(file_string: str) -> dict:
    """Parse mixed product wave function information.

    :param str file_string: Input string
    :return dict data: Matched data
    """
    wf_info_keys = [
        'Maximal number of MT wavefunctions per atom:',
        'Total number of MT wavefunctions:',
        'Maximal number of PW wavefunctions:',
        'Total number of mixed-product wavefunctions:'
    ]

    return parse_values_regex(file_string, wf_info_keys)


def parse_frequency_grid_info(file_string: str) -> dict:
    """Parse frequency grid information.

    :param str file_string: Input string
    :return dict data: Matched data
    """
    fgrid_keys = ['Type: < fgrid >',
                  'Frequency axis: < fconv >',
                  'Number of frequencies: < nomeg >',
                  'Cutoff frequency: < freqmax >'
                  ]

    return parse_values_regex(file_string, fgrid_keys)


def parse_frequency_grid(file_string: str, n_points: int) -> np.ndarray:
    """Parse the frequency grid and weights used with GW

    :param str file_string: Input string
    :param int n_points: Number of frequency grid points
    :return np.ndarray grid_and_weights: Frequency grid and weights in grid[0, :] and grid[1, :],
    respectively
    """
    index = file_string.find('frequency list: < #    freqs    weight >')
    frequency_lines = file_string[index:].split('\n')

    grid_and_weights = np.empty(shape=(2, n_points))
    for i in range(0, n_points):
        index, frequency, weight = frequency_lines[i + 1].split()
        grid_and_weights[0:2, i] = np.array([float(frequency), float(weight)])

    return grid_and_weights


def parse_ks_eigenstates(file_string: str) -> dict:
    """Parse information on the KS eigenstates used.
    Note, final keys will have the trailing whitespace stripped.

    :param str file_string: Input string
    :return dict data: Matched data
    """

    # Trailing whitespace in some instances required for match
    ks_eigenstates_keys = ['Maximum number of LAPW states:',
                           'Minimal number of LAPW states:',
                           '- total KS',
                           '- occupied',
                           '- unoccupied ',
                           '- dielectric function',
                           '- self energy',
                           'Energy of the highest unoccupied state: ',
                           'Number of valence electrons:',
                           'Number of valence electrons treated in GW: '
                           ]

    data = parse_values_regex(file_string, ks_eigenstates_keys)

    # Prepend keys that lack context, such as '- total KS'
    modified_data = {}
    prepend_str = "Number of states used in GW"

    for key, value in data.items():
        if key[0] == '-':
            new_key = prepend_str + " " + key.rstrip().rstrip(':')
        else:
            new_key = key.rstrip().rstrip(':')
        modified_data[new_key] = value

    return modified_data


def parse_n_q_point_cycles(file_string: str) -> int:
    """Get the maximum number of q iterations performed.

    :param str file_string: Input string
    :return int n_q_cycles:  maximum nunber of q iterations performed
    """
    matches = re.findall('\\(task_gw\\): q-point cycle, iq =' + '(.+?)\n',
                         file_string)
    n_q_cycles = max([int(string.strip()) for string in matches])
    return n_q_cycles


def extract_kpoint(file_string: str) -> dict:
    """Parse the substring of the form:

     at k =    0.000   0.500   0.500 ik =     3

    returning a dictionary of the form:

     {'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
      'CBm': {'k_point': [0.0, 0.5, 0.5], 'ik': 3}
      }

    :param str file_string: Input string
    :return dict k_data: Matched data, of the form documented
    """

    def parse_k_match(file_string: str, key: str) -> dict:
        data = {}

        try:
            match = re.search(key + '(.+?)\n', file_string)
            k_point_and_index = match.group(1).split()
            data = {'k_point': [float(k) for k in k_point_and_index[:3]],
                    'ik': int(k_point_and_index[-1])}

        except AttributeError:
            raise AttributeError("extract_kpoint. Did not find the key", match)

        return data

    k_data = parse_k_match(file_string, 'at k      =    ')

    return {'VBM': k_data, 'CBm': k_data}


def extract_kpoints(file_string: str) -> dict:
    """Parse the substring of the form:

     at k(VBM) =    0.000   0.500   0.500 ik =     3
        k(CBm) =    0.000   0.000   0.000 ik =     1

    returning a dictionary of the form:

     {'VBM': {'k_point': [0.0, 0.5, 0.5], 'ik': 3},
      'CBm': {'k_point': [0.0, 0.0, 0.0], 'ik': 1}
      }

    :param str file_string: Input string
    :return dict k_data: Matched data, of the form documented
    """

    def parse_k_match(file_string: str, key: str) -> dict:
        data = {}

        match_key_to_parser_key = {
            'at k\\(VBM\\) = ': 'VBM',
            'k\\(CBm\\) = ': 'CBm'
        }

        try:
            match = re.search(key + '(.+?)\n', file_string)
            parser_key = match_key_to_parser_key[key].replace('\\', "")
            k_point_and_index = match.group(1).split()
            data[parser_key] = {'k_point': [float(k) for k in k_point_and_index[:3]],
                                'ik': int(k_point_and_index[-1])}

        except AttributeError:
            print("extract_kpoints. Did not find the key", match)

        return data

    k_data = parse_k_match(file_string, 'at k\\(VBM\\) = ')
    k_data2 = parse_k_match(file_string, 'k\\(CBm\\) = ')

    return {** k_data, **k_data2}


def parse_band_structure_info(file_string: str, bs_type: str) -> dict:
    """Parse KS or GW band structure information.

    This routine assumes that the KS band structure info will ALWAYS appear
    before the GW band structure info.

    Two situations can occur.

    * 1. Indirect bandgap:

     Indirect BandGap (eV):                    3.3206
     at k(VBM) =    0.000   0.500   0.500 ik =     3
        k(CBm) =    0.000   0.000   0.000 ik =     1
     Direct Bandgap at k(VBM) (eV):            3.7482
     Direct Bandgap at k(CBm) (eV):            3.8653

    * 2. Direct bandgap:

      Direct BandGap (eV):                      2.3903
      at k      =    0.000   0.000   0.000 ik =     1

    :param str bs_type: Band structure type to parse. Either 'ks' or 'gw'
    :param str file_string: Input string
    :return dict k_data: Matched data
    """
    if bs_type == 'ks':
        # Parse first instance of each key, exploiting that Kohn-Sham band structure
        # comes before G0W0 band structure. This ASSUMES fixed structure to the file
        pass

    elif bs_type == 'gw':
        # Find G0W0 band structure in the file, then start parsing from there
        gw_header = ' G0W0 band structure '
        index = file_string.find('G0W0 band structure ')
        file_string = file_string[index:]

    else:
        sys.exit("bs_type must be 'ks' or 'gw'")

    # Indirect BandGap may not be present
    band_structure_keys = ['Fermi energy:',
                           'Energy range:',
                           'Band index of VBM:',
                           'Band index of CBm:']

    data = parse_values_regex(file_string, band_structure_keys)

    # Only present if there's an indirect gap
    indirect_gap = parse_value_regex(file_string,
                                     'Indirect BandGap \\(eV\\):',
                                     silent_key_error=True)

    if indirect_gap:
        direct_gap_keys = [
            'Direct Bandgap at k\\(VBM\\) \\(eV\\):',
            'Direct Bandgap at k\\(CBm\\) \\(eV\\):'
        ]
        data.update(indirect_gap)
        data.update(parse_values_regex(file_string, direct_gap_keys))
        k_point_data = extract_kpoints(file_string)

    else:
        data.update(parse_value_regex(file_string, 'Direct BandGap \\(eV\\):'))
        k_point_data = extract_kpoint(file_string)

    return {**data, **k_point_data}


@accept_file_name
def parse_gw_info(file_string: str) -> dict:
    """Parse data from GW_INFO.OUT.

    Timings are not parsed because it's more convenient for regression-testing.

    :param str file_string: Parsed file string.
    :return: dict data: dictionary of parsed data.
    """
    data = {}
    data['correlation_self_energy_parameters'] = parse_correlation_self_energy_params(file_string)
    data['mixed_product_basis_parameters'] = parse_mixed_product_params(file_string)
    data['bare_coulomb_potential_parameters'] = parse_bare_coulomb_potential_params(file_string)
    data['screened_coulomb_potential'] =  match_current_return_line_n(file_string, 'Screened Coulomb potential:').strip()
    data['core_electrons_treatment'] = match_current_return_line_n(file_string, 'Core electrons treatment:').strip()
    data['qp_interval'] = parse_value_regex(file_string, 'Interval of quasiparticle states \\(ibgw, nbgw\\):')\
        ['Interval of quasiparticle states (ibgw, nbgw)']
    data['n_empty'] = parse_value_regex(file_string, 'Number of empty states \\(GW\\):')['Number of empty states (GW)']
    data['q_grid'] = parse_value_regex(file_string, 'k/q-points grid:')['k/q-points grid']
    data['mixed_product_wf_info'] = parse_mixed_product_wf_info(file_string)
    data['frequency_grid'] = parse_frequency_grid_info(file_string)
    n_freq_points = data['frequency_grid']['Number of frequencies: < nomeg >']
    data['frequency_grid']['frequencies_weights'] = parse_frequency_grid(file_string, n_freq_points)
    data['ks_eigenstates_summary'] = parse_ks_eigenstates(file_string)
    data['ks_band_structure_summary'] = parse_band_structure_info(file_string, 'ks')
    data['n_q_cycles'] = parse_n_q_point_cycles(file_string)
    data['g0w0_band_structure_summary'] = parse_band_structure_info(file_string, 'gw')
    return data


def extract_gw_timings_as_list(file_string: str) -> Union[List[str], None]:
    """Extract GW timing string block as a list.

    Utilises the fact that timings are returned at the end of GW_INFO.OUT

    :param str file_string: Parsed file string
    :param list timings: GW timings, with each element storing a
    line of timings as a string.
    """
    file_list = file_string.split('\n')
    for i, line in enumerate(reversed(file_list)):
        if 'GW timing info (seconds)' in line:
            index = i - 2
            return file_list[-index:]

    return None


def parse_gw_timings(file_string: str) -> dict:
    """Parse timings returned by the GW method.

    Assumptions:
     * Initial timing value is 'Initialization'
     * Final timing value is 'Total'

    Returns a dictionary with each time subheading storing
    a dictionary of timing breakdowns.

    If the subheading has a time associated with it, that is stored
    in the breakdowns with the same key. If it has no associated timing,
    the value is None.

    For example:

    'Subroutines': {'Subroutines': None,
                    'calcpmat': 5.12,
                    'calcbarcmb': 5.65,
                    'BZ integration weights': 18.35
                    }

    :param str file_string: Parsed file string
    :return: dict data: dictionary of parsed timings
    """
    timings = extract_gw_timings_as_list(file_string)

    # Parse and store nested timings
    data = {}
    component_time = {}
    root_key = 'Initialization'
    initial_key = root_key

    for line in timings:

        key = line.strip().split(':')[0].rstrip()
        if len(key) == 0: continue
        is_root_key = key[0] != '-'

        if is_root_key and (key != initial_key):
            data[root_key] = copy.deepcopy(component_time)
            component_time.clear()
            root_key = key

        time_str = line.strip().split(':')[-1]
        component_time[key.strip('-').strip()] = float(
            time_str) if can_be_float(time_str) else None

    # Store last value, Total
    data[root_key] = {root_key: float(time_str)}

    # Remove superfluous key from table formatting
    arb_str = '_________________________________________________________'
    del data[arb_str]

    return data



"""Test XML utilities.
"""

from excitingtools.input.xml_utils import line_reformatter


def test_line_reformatter_long_ending():
    """Test reformatting of XML as a string works."""
    input_str = (
        '<groundstate do="fromscratch" ngridk="6 6 6" nosource="false" '
        'rgkmax="8.0" tforce="true" vkloff="0 0 0" xctype="GGA_PBE_SOL"> '
        '</groundstate>'
    )
    pretty_input_str = line_reformatter(input_str)
    # Note, whitespace is important
    reference = """<groundstate
   do="fromscratch"
   ngridk="6 6 6"
   nosource="false"
   rgkmax="8.0"
   tforce="true"
   vkloff="0 0 0"
   xctype="GGA_PBE_SOL">
</groundstate>"""
    assert reference == pretty_input_str


def test_line_reformatter_short_ending():
    """Test reformatting of XML as a string works for xml strings
    with a different (short) ending."""
    input_str = (
        '<groundstate do="fromscratch" ngridk="6 6 6" nosource="false" '
        'rgkmax="8.0" tforce="true" vkloff="0 0 0" xctype="GGA_PBE_SOL"/> '
    )
    pretty_input_str = line_reformatter(input_str)
    # Note, whitespace is important
    reference = """<groundstate
   do="fromscratch"
   ngridk="6 6 6"
   nosource="false"
   rgkmax="8.0"
   tforce="true"
   vkloff="0 0 0"
   xctype="GGA_PBE_SOL"/>"""
    assert reference == pretty_input_str


def test_line_reformatter_no_closing():
    """Test reformatting of XML as a string works for xml strings
    with no closing at the end."""
    input_str = (
        '<groundstate do="fromscratch" ngridk="6 6 6" nosource="false" '
        'rgkmax="8.0" tforce="true" vkloff="0 0 0" xctype="GGA_PBE_SOL"> '
    )
    pretty_input_str = line_reformatter(input_str)
    # Note, whitespace is important
    reference = """<groundstate
   do="fromscratch"
   ngridk="6 6 6"
   nosource="false"
   rgkmax="8.0"
   tforce="true"
   vkloff="0 0 0"
   xctype="GGA_PBE_SOL">"""
    assert reference == pretty_input_str



"""Test exiting input base class and its methods.

NOTE:
All attribute tests should assert on the XML tree content's as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.
"""

import pathlib

from excitingtools.input.base_class import query_exciting_version


def test_query_exciting_version(tmp_path):
    """
    Test querying the exciting version.
    """
    version_inc_contents = """#define GITHASH "1a2087b0775a87059d53"
#define GITHASH2 "5d01a5475a10f00d0ad7"
#define COMPILERVERSION "GNU Fortran (MacPorts gcc9 9.3.0_4) 9.3.0"
#define VERSIONFROMDATE /21,12,01/
    """

    # Mock the version.inc file, and prepended path
    exciting_root = pathlib.Path(tmp_path)
    src = exciting_root / "src"
    src.mkdir()
    assert exciting_root.is_dir(), "exciting_root tmp_path directory does not exist"

    version_inc = exciting_root / "src" / "version.inc"
    version_inc.write_text(version_inc_contents)
    assert version_inc.exists(), "version.inc does not exist"

    versioning: dict = query_exciting_version(exciting_root)
    assert set(versioning.keys()) == {'compiler', 'git_hash'}, (
        "Expect `query_exciting_version` to return compiler used "
        "for last build, and exciting git hash")

    assert versioning['compiler'] == 'GNU Fortran (MacPorts gcc9 9.3.0_4) 9.3.0'
    assert versioning['git_hash'] == '1a2087b0775a87059d535d01a5475a10f00d0ad7'



"""
Tests for functions in math_utils.py
"""
import numpy as np

from excitingtools.math.math_utils import unit_vector


def test_unit_vector():
    """Test unit_vector returns a vector with norm of one."""
    x = np.array([0.23, 0.32, 1.394, 99.])
    norm_x = np.linalg.norm(unit_vector(x))
    assert np.isclose(norm_x, 1., atol=1.e-8)



"""Test composition of an exciting input XML.

TODO(Fab/Alex/Dan) Issue 117. Would be nice to assert that the output is valid
    XML * https://lxml.de/validation.html
Also see: https://xmlschema.readthedocs.io/en/latest/usage.html#xsd-declarations

NOTE:
All attribute tests should assert on the XML tree content's as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.
"""

import pytest

from excitingtools.input.input_classes import ExcitingGroundStateInput, ExcitingXSInput, ExcitingInputXML
from excitingtools.input.structure import ExcitingStructure


@pytest.fixture
def exciting_structure() -> ExcitingStructure:
    """ Initialise an exciting structure. """
    cubic_lattice = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    arbitrary_atoms = [{'species': 'Li', 'position': [0.0, 0.0, 0.0]},
                       {'species': 'Li', 'position': [1.0, 0.0, 0.0]},
                       {'species': 'F', 'position': [2.0, 0.0, 0.0]}]

    return ExcitingStructure(arbitrary_atoms, cubic_lattice, '.')


@pytest.fixture
def exciting_input_xml(exciting_structure: ExcitingStructure) -> ExcitingInputXML:
    """ Initialises a complete input file. """
    ground_state = ExcitingGroundStateInput(
        rgkmax=8.0,
        do="fromscratch",
        ngridk=[6, 6, 6],
        xctype="GGA_PBE_SOL",
        vkloff=[0, 0, 0],
        tforce=True,
        nosource=False
    )

    bse_attributes = {'bsetype': 'singlet', 'xas': True}
    energywindow_attributes = {'intv': [5.8, 8.3], 'points': 5000}
    screening_attributes = {'screentype': 'full', 'nempty': 15}
    plan_input = ['screen', 'bse']
    qpointset_input = [[0, 0, 0], [0.5, 0.5, 0.5]]
    xs = ExcitingXSInput(xstype="BSE", broad=0.32, ngridk=[8, 8, 8],
                         BSE=bse_attributes,
                         energywindow=energywindow_attributes,
                         screening=screening_attributes,
                         qpointset=qpointset_input,
                         plan=plan_input)

    return ExcitingInputXML(sharedfs=True, structure=exciting_structure, title='Test Case', groundstate=ground_state,
                            xs=xs)


def test_exciting_input_xml_structure_and_gs_and_xs(exciting_input_xml: ExcitingInputXML):
    """Test the XML created for a ground state input is valid.
    Test SubTree composition using only mandatory attributes for each XML subtree.
    """
    input_xml_tree = exciting_input_xml.to_xml()

    assert input_xml_tree.tag == 'input'
    assert input_xml_tree.keys() == ['sharedfs']

    subelements = list(input_xml_tree)
    assert len(subelements) == 4

    title_xml = subelements[0]
    assert title_xml.tag == 'title'
    assert title_xml.keys() == []
    assert title_xml.text == 'Test Case'

    structure_xml = subelements[1]
    assert structure_xml.tag == 'structure'
    assert structure_xml.keys() == ['speciespath']
    assert len(list(structure_xml)) == 3

    groundstate_xml = subelements[2]
    assert groundstate_xml.tag == 'groundstate'
    assert groundstate_xml.text == ' '
    assert groundstate_xml.keys() == \
           ['rgkmax', 'do', 'ngridk', 'xctype', 'vkloff', 'tforce', 'nosource']
    assert groundstate_xml.get('rgkmax') == "8.0"
    assert groundstate_xml.get('do') == "fromscratch"
    assert groundstate_xml.get('ngridk') == "6 6 6"
    assert groundstate_xml.get('xctype') == "GGA_PBE_SOL"
    assert groundstate_xml.get('vkloff') == "0 0 0"
    assert groundstate_xml.get('tforce') == "true"
    assert groundstate_xml.get('nosource') == "false"

    xs_xml = subelements[3]
    assert xs_xml.tag == 'xs'
    assert set(xs_xml.keys()) == {'broad', 'ngridk', 'xstype'}
    assert xs_xml.get('broad') == '0.32'
    assert xs_xml.get('ngridk') == '8 8 8'
    assert xs_xml.get('xstype') == 'BSE'

    xs_subelements = list(xs_xml)
    assert len(xs_subelements) == 5
    valid_tags = {"screening", "BSE", "energywindow", "qpointset", "plan"}
    assert valid_tags == set(xs_subelement.tag for xs_subelement in xs_subelements)

    screening_xml = xs_xml.find("screening")
    assert screening_xml.tag == "screening"
    assert screening_xml.keys() == ['screentype', 'nempty']
    assert screening_xml.get('screentype') == 'full'
    assert screening_xml.get('nempty') == '15'

    bse_xml = xs_xml.find("BSE")
    assert bse_xml.tag == 'BSE'
    assert bse_xml.keys() == ['bsetype', 'xas']
    assert bse_xml.get('bsetype') == 'singlet'
    assert bse_xml.get('xas') == 'true'

    energywindow_xml = xs_xml.find("energywindow")
    assert energywindow_xml.tag == "energywindow"
    assert energywindow_xml.keys() == ['intv', 'points']
    assert energywindow_xml.get('intv') == '5.8 8.3'
    assert energywindow_xml.get('points') == '5000'

    qpointset_xml = xs_xml.find("qpointset")
    assert qpointset_xml.tag == "qpointset"
    assert qpointset_xml.items() == []
    qpoints = list(qpointset_xml)
    assert len(qpoints) == 2
    assert qpoints[0].tag == 'qpoint'
    assert qpoints[0].items() == []
    valid_qpoints = {'0 0 0', '0.5 0.5 0.5'}
    assert qpoints[0].text in valid_qpoints
    valid_qpoints.discard(qpoints[0].text)
    assert qpoints[1].text in valid_qpoints

    plan_xml = xs_xml.find("plan")
    assert plan_xml.tag == "plan"
    assert plan_xml.items() == []
    doonlys = list(plan_xml)
    assert len(doonlys) == 2
    assert doonlys[0].tag == 'doonly'
    assert doonlys[0].items() == [('task', 'screen')]
    assert doonlys[1].tag == 'doonly'
    assert doonlys[1].items() == [('task', 'bse')]


def test_attribute_modification(exciting_input_xml: ExcitingInputXML):
    """Test the XML created for a ground state input is valid.
    Test SubTree composition using only mandatory attributes for each XML subtree.
    """
    exciting_input_xml.set_title("New Test Case")
    exciting_input_xml.structure.crystal_properties.scale = 2.3
    exciting_input_xml.groundstate.rgkmax = 9.0
    exciting_input_xml.xs.energywindow.points = 4000
    input_xml_tree = exciting_input_xml.to_xml()

    subelements = list(input_xml_tree)
    assert len(subelements) == 4

    title_xml = subelements[0]
    assert title_xml.tag == 'title'
    assert title_xml.text == 'New Test Case'

    structure_xml = subelements[1]
    assert structure_xml[0].get("scale") == "2.3"

    groundstate_xml = subelements[2]
    assert groundstate_xml.get('rgkmax') == "9.0"

    xs_xml = subelements[3]
    xs_subelements = list(xs_xml)
    assert len(xs_subelements) == 5

    energywindow_xml = xs_xml.find("energywindow")
    assert energywindow_xml.get('points') == '4000'


def test_as_dict(exciting_input_xml: ExcitingInputXML, mock_env_jobflow_missing):
    dict_representation = exciting_input_xml.as_dict()
    assert set(dict_representation.keys()) == {"xml_string"}
    # check only that the xml string starts with the correct first lines:
    assert dict_representation["xml_string"].startswith('<?xml version="1.0" ?>\n<input sharedfs="true">\n\t'
                                                        '<title>Test Case</title>\n\t<structure')


def test_as_dict_jobflow(exciting_input_xml: ExcitingInputXML, mock_env_jobflow):
    dict_representation = exciting_input_xml.as_dict()
    xml_string = dict_representation.pop("xml_string")
    assert dict_representation == {'@class': 'ExcitingInputXML',
                                   '@module': 'excitingtools.input.input_classes'}
    # check only that the xml string starts with the correct first lines:
    assert xml_string.startswith('<?xml version="1.0" ?>\n<input sharedfs="true">\n\t'
                                 '<title>Test Case</title>\n\t<structure')


def test_from_dict(exciting_input_xml: ExcitingInputXML, mock_env_jobflow_missing):
    new_input_xml = ExcitingInputXML.from_dict(exciting_input_xml.as_dict())
    assert new_input_xml.title.title == "Test Case"  # pylint: disable=no-member
    assert new_input_xml.groundstate.ngridk == [6, 6, 6]  # pylint: disable=no-member


def test_from_dict_jobflow(exciting_input_xml: ExcitingInputXML, mock_env_jobflow):
    new_input_xml = ExcitingInputXML.from_dict(exciting_input_xml.as_dict())
    assert new_input_xml.title.title == "Test Case"  # pylint: disable=no-member


def test_missing_structure():
    with pytest.raises(ValueError, match="Missing mandatory arguments: {'structure'}"):
        ExcitingInputXML(title="Test Case", groundstate=ExcitingGroundStateInput())


def test_from_gs_dict(exciting_structure):
    input_xml = ExcitingInputXML(structure=exciting_structure, title="Test Case",
                                 groundstate={"rgkmax": 7.0})
    assert input_xml.title.title == "Test Case"  # pylint: disable=no-member
    assert input_xml.groundstate.rgkmax == 7.0  # pylint: disable=no-member
    assert input_xml.structure.speciespath == "."  # pylint: disable=no-member



"""Decorators and wrappers for parser functions.
"""
import pathlib
import xml.etree.ElementTree as ET
from typing import Callable, Union


def return_file_string(file_name: Union[str, pathlib.Path]) -> str:
    """ Given a file name, return the file contents as a string.

    :param file_name: File name.
    :return file_string: File contents string.
    """
    file_name_ = file_name

    if isinstance(file_name_, str):
        file_name_ = pathlib.Path(file_name_)

    if not file_name_.exists():
        raise FileNotFoundError(f'{file_name_} not found')

    return file_name_.read_text()


def file_handler(file_name: Union[str, pathlib.Path], parser_func: Callable[[str], dict]) -> dict:
    """Provide a wrapper for file IO.

    :param file_name: File name or Path object
    :param Callable[[str], dict] parser_func: Parser function, which expects a parsed
     string as its only input and returns a dictionary.
    :return: dict data: Dictionary of parsed data, with values converted from strings.
    """
    file_string = return_file_string(file_name)
    return parser_func(file_string)


def accept_file_name(parser: Callable):
    """ Decorate parsers that accept string contents, such that they take file names instead.
    """

    def modified_func(file_name: Union[str, pathlib.Path]):
        """ Wrapper.
        param: file_name: File name.
        """
        file_string = return_file_string(file_name)
        return parser(file_string)
    return modified_func


def xml_root(func: Callable):
    """ Decorate XML parsers, enabling the developer to pass
    an XML file name, XML string or ElementTree.Element as input
    and return the XML root.
    """
    function_selection = {type(None): lambda x, y: func(x), str: lambda x, y: func(x, y)}

    def modified_func(input: str, tag: str = None):
        # Element
        if isinstance(input, ET.Element):
            return function_selection[type(tag)](input, tag)

        # File name
        try:
            tree = ET.parse(input)
            root = tree.getroot()
            return function_selection[type(tag)](root, tag)
        except (FileNotFoundError, OSError):
            pass

        # XML string
        try:
            root = ET.fromstring(input)
            return function_selection[type(tag)](root, tag)
        except (ET.ParseError, TypeError):
            raise ValueError(f'Input string neither an XML file, '
                             f'nor valid XML: {input}')

    return modified_func



"""Parsers for exciting properties.
"""
import xml.etree.cElementTree as ET
from xml.etree.ElementTree import ParseError
import numpy as np
import os
from typing import Dict

from excitingtools.parser_utils.parser_decorators import xml_root


def parse_plot_3d(name: str) -> dict:
    """
    Parser for RHO3D.xml, VCL3D.xml, VXC3D.xml, WF3D.xml, ELF3D.xml, EF3D.xmlit

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name)
    plot_3d = {"title": root.find("title").text}
    grid = root.find("grid").attrib
    axis = []
    for ax in root.find("grid").findall("axis"):
        axis.append(ax.attrib)
    grid["axis"] = {}
    for item in axis:
        name = item["name"]
        grid["axis"][name] = item
    grid["value"] = root.find("grid").find("value").attrib
    plot_3d["grid"] = grid

    func_node = root.find("function")
    function = root.find("function").attrib
    row0 = []
    for node in func_node.findall("row"):
        row = node.attrib
        row1 = []
        for nod in node:
            row2 = nod.attrib
            row2["data"] = nod.text.split()
            row1.append(row2)
        row["row"] = {}
        for item in row1:
            name = item["index"]
            row["row"][name] = item
        row0.append(row)
    function["row"] = {}
    for item in row0:
        name = item["index"]
        function["row"][name] = item
    plot_3d["function"] = function

    return plot_3d


def parse_lsj(name: str) -> dict:
    """
    Parser for LSJ.xml

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    LSJ = {}
    species = []
    for node in root.findall("species"):
        spec = node.attrib
        atom = []
        for nod in node.findall("atom"):
            at = nod.attrib
            at["L"] = nod.find("L").text.split()
            at["S"] = nod.find("S").text.split()
            at["J"] = nod.find("J").text.split()
            atom.append(at)
        spec["atom"] = {}
        for item in atom:
            name = item["n"]
            spec["atom"][name] = item
        species.append(spec)
    LSJ["species"] = {}
    for item in species:
        name = item['n']
        LSJ['species'][name] = item

    return LSJ


def parse_efg(name: str) -> dict:
    """
    Parser for EFG.xml

    Returns a dictionary of the form:
      data = {'species1': {'chemicalSymbol': chemicalSymbol,
                          'atom1': { 'trace': trace,
                                     'efg: efg,
                                     'eigenvalues': eigenvalues
                                   },
                          'atom2': {},
                          }
              'species2':...
            }

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    data = {}

    for species in root.findall('species'):
        species_key = species.tag + str(species.attrib['n'])
        data[species_key] = {
            'chemicalSymbol': species.attrib['chemicalSymbol']
        }

        for atom in species.findall('atom'):
            atom_key = atom.tag + atom.attrib['n']

            for efg_tensor in atom.findall("EFG-tensor"):
                efg = np.empty(shape=(3, 3))

                for i, line in enumerate(efg_tensor.findall("line")):
                    efg[i, :] = [float(x) for x in line.text.split()]

                for eigenvalues in efg_tensor.findall("eigenvalues"):
                    eig = [float(e) for e in eigenvalues.text.split()]

            data[species_key][atom_key] = {
                'trace': float(efg_tensor.attrib['trace']),
                'efg': efg,
                'eigenvalues': eig
            }
    return data


def parse_mossbauer(name: str) -> dict:
    """
    Parser for mossbauer.xml

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    mossbauer = {}
    species = []
    for node in root.findall("species"):
        spec = node.attrib
        atom = []
        for nod in node:
            atom.append(nod.attrib)
        spec["atom"] = {}
        for item in atom:
            name = item["n"]
            spec["atom"][name] = item
        species.append(spec)
    mossbauer["species"] = {}
    for item in species:
        name = item["n"]
        mossbauer["species"][name] = item

    return mossbauer


def parse_expiqr(name: str) -> dict:
    """
    Parser for expiqr.xml

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    expiqr = {}
    expiqr["q-vector"] = root.find("q-vector").attrib
    kgrid = {}
    for k in root.find("k-grid"):
        kgrid = k.attrib
        states = []
        for st in k.findall("state"):
            state = st.attrib
            states.append(state)
            statesj = []
            for s in st.findall("state"):
                statej = s.attrib
                statesj.append(statej)
            state["state"] = {}
            for item in statesj:
                name = item["j"]
                state["state"][name] = item
        kgrid["state"] = {}
        for item in states:
            name = item["i"]
            kgrid["state"][name] = item
    expiqr["k-grid"] = kgrid
    return expiqr


def parse_effmass(name: str) -> dict:
    """
    Parser for effmass.xml

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    effmass = {}
    effmass["k-point"] = root.find("k-point").attrib
    state = []
    for node in root.findall("state"):
        st = node.attrib

        evdk = node.find("evdk").attrib
        matrix1 = []
        for line in node.find("evdk").findall("line"):
            matrix1.append(line.text.split())
        evdk["evdk_matrix"] = matrix1

        emt = node.find("emt").attrib
        matrix2 = []
        for line in node.find("emt").findall("line"):
            matrix2.append(line.text.split())
        emt["emt_matrix"] = matrix2

        st["evdk"] = evdk
        st["emt"] = emt
        state.append(st)

    effmass["state"] = {}
    for item in state:
        name = item["n"]
        effmass["state"][name] = item

    return effmass


# TODO(Hannah). Issue 138. Ensure test cases work with `parse_bandstructure` and remove `parse_bandstructure_depreciated`
# This parser is depreciated. Please do not use.
def parse_bandstructure_depreciated(name: str) -> dict:
    """
    Parser for bandstructure.xml.

    Used for parsing in the test framework, as returns a dict.

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    bandstructure = {}
    bandstructure["title"] = root.find("title").text
    bands = []
    for node in root.findall("band"):
        band = node.attrib
        points = []
        for nod in node.findall("point"):
            point = nod.attrib
            points.append(point)
        bands.append(band)
        band["point"] = {}
        i = 1
        for item in points:
            name = str(i)
            band["point"][name] = item
            i = i + 1
    bandstructure["band"] = {}
    j = 1
    for item in bands:
        name = str(j)
        bandstructure["band"][name] = item
        j = j + 1
    return bandstructure


@xml_root
def parse_band_structure_xml(root) -> dict:
    """ Parse KS band structure from bandstructure.xml.

    :param root: Band structure XML file name, XML string or ElementTree.Element as input.
    :return: Band data
    """
    # Split band structure file contents: title, bands and vertices
    bs_xml: Dict[str, list] = {'title': [], 'band': [], 'vertex': []}

    for item in list(root):
        try:
            bs_xml[item.tag].append(item)
        except KeyError:
            raise KeyError(f'Element tag {item.tag} requires implementing in band structure parser')

    n_bands = len(bs_xml['band'])
    first_band = bs_xml['band'][0]
    n_kpts = len(list(first_band))

    # Same set of flattened k-points, per band - so parse once
    k_points_along_band = np.array([point.get('distance') for point in list(first_band)], dtype=float)

    # Read E(k), per band
    band_energies = np.empty(shape=(n_kpts, n_bands))
    for ib, band in enumerate(bs_xml['band']):
        for ik, point in enumerate(list(band)):
            band_energies[ik, ib] = point.get('eval')

    vertices = []
    for element in bs_xml['vertex']:
        vertices.append({'distance': float(element.get('distance')),
                         'label': element.get('label'),
                         'coord':  [float(x) for x in element.get('coord').split()]})

    return {'title': bs_xml['title'], 'n_kpts': n_kpts, 'n_bands': n_bands,
            'k_points_along_band': k_points_along_band,
            'band_energies': band_energies, 'vertices': vertices}


def parse_band_structure_dat(name: str) -> dict:
    """Parser for bandstructure.dat

    :param str name: File name
    :return dict output: Parsed data
    """
    bs_dat = np.loadtxt(name)
    with open(name) as f:
        header = f.readline()

    n_kpts = int(header.split()[3])
    n_bands = int(header.split()[2])
    dimensions = 3

    k_points = np.empty(shape=(n_kpts, dimensions))
    flattened_k_points = np.empty(n_kpts)
    for i in range(n_kpts):
        k_points[i] = np.array([k for k in bs_dat[i, 2:5]])
        flattened_k_points[i] = bs_dat[i, 5]

    band_energies = np.reshape(bs_dat[:, 6], (n_kpts, n_bands), order='F')

    return {
            'n_kpts': n_kpts,
            'n_bands': n_bands,
            'k_points': k_points,
            'flattened_k_points': flattened_k_points,
            'band_energies': band_energies
        }


def parse_dos(name: str) -> dict:
    """
    Parser for dos.xml

    :param str name: File name
    :return dict output: Parsed data
    """
    root = ET.parse(name).getroot()
    dos = {}
    dos["title"] = root.find("title").text
    totaldos = root.find("totaldos").attrib
    dos["totaldos"] = totaldos
    diagram = root.find("totaldos").find("diagram").attrib
    dos["totaldos"]["diagram"] = diagram
    points = []
    for node in root.find("totaldos").find("diagram").findall("point"):
        point = node.attrib
        points.append(point)
    dos["totaldos"]["diagram"]["point"] = {}
    i = 1
    for item in points:
        name = str(i)
        dos["totaldos"]["diagram"]["point"][name] = item
        i = i + 1
    return dos


@xml_root
def parse_charge_density(root) -> np.ndarray:
    """ Parse charge density from RHO1D.xml file.

    `axis` and `vertex` sub-trees ignored in the parsing.

    :param root: XML file name, XML string or ElementTree.Element as input.
    :return: Numpy array containing rho[:, 1] = distance and rho[:, 2] = density.
    """
    function_points = root.find('grid').find('function')
    rho = np.empty(shape=(len(function_points), 2))
    for i, point in enumerate(function_points):
        rho[i, :] = [point.attrib['distance'], float(point.attrib['value'])]
    return rho


def parse_kerr(name: str) -> dict:
    """
    Parser for KERR.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name, skip_header=1)
    except:
        raise ParseError

    out = {'energy': data[:, 0], 're': data[:, 1], 'im': data[:, 2]}

    return out


def parse_epsilon(name: str) -> dict:
    """
    Parser for EPSILON_ij.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name, skip_header=1)
    except:
        raise ParseError
    out = {'energy': data[:, 0], 're': data[:, 1], 'im': data[:, 2]}
    return out


def parse_chi(name: str) -> dict:
    """
    Parser for CHI_111.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name, skip_header=1)
    except:
        raise ParseError
    out = {
        'energy': data[:, 0],
        're': data[:, 1],
        'im': data[:, 2],
        'modulus': data[:, 3]
    }
    return out


def parse_elnes(name: str) -> dict:
    """
    Parser for ELNES.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name)
    except:
        raise ParseError
    out = {'energy': data[:, 0], 'elnes': data[:, 1]}
    return out


def parse_seebeck(name: str) -> dict:
    """
    Parser for SEEBECK_11.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name)
    except:
        raise ParseError
    out = {
        'temperature': data[:, 0],
        'mu': data[:, 1],
        're': data[:, 2],
        'im': data[:, 3]
    }

    return out


def parse_ldos(name: str) -> dict:
    """
    Parser for ldos.out

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name)
    except:
        raise ParseError
    out = {'energy': data[:, 0], 'ldos': data[:, 1]}

    return out


def parse_band_edges(name: str) -> dict:
    """
    Parser for band_edges.out

    Keys
    * c_axis corresponds to the linear grid along the magnitude
      of the c vector of the unit cell.
    * VBM = Valence band maximum
    * CBm = Conduction band minimum

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name)
    except:
        raise ParseError
    out = {'c_axis': data[:, 0], 'VBM': data[:, 1], 'CBm': data[:, 2]}

    return out


def parse_spintext(name: str) -> dict:
    """
    Parse spintext.xml

    TODO(Bene) Issue 87 Refactor to return a dict

    Each element of the list contains a dict with keys:
     ['ist', 'k-point', 'spin', 'energy']

    :param str name: Path to the spintext.xml that will be parsed
    :return dict spintext: List that holds the parsed spintexture.xml
    """
    # parse file
    file_name = 'spintext.xml'
    if name.split('/')[-1] != file_name:
        name = os.path.join(name, file_name)

    tree_spin = ET.parse(name)
    root_spin = tree_spin.getroot()

    spintext = {}
    i = 0
    for band in root_spin.findall("band"):
        k_point = []
        spin = []
        energy = []

        for val in band.findall("k-point"):
            k_point.append([float(k) for k in val.attrib["vec"].split()])
            spin.append([float(s) for s in val.attrib["spin"].split()])
            energy.append(float(val.attrib["energy"]))

        spintext[str(i)] = {
            "ist": int(band.attrib["ist"]),
            "k-point": k_point,
            "spin": spin,
            "energy": energy
        }
        i += 1

    return spintext


def parse_polarization(name: str) -> dict:
    """
    Parser for POLARIZATION.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    file = open(name)
    lines = []
    for i in range(len(open(name).readlines())):
        line = next(file)
        if '#' not in line:
            lines.append(line.split())
    polarization = {
        'total': lines[0],
        'electronic': lines[1],
        'ionic': lines[2]
    }
    return polarization


def parse_tdos_wannier(name: str) -> dict:
    """
    Parser for TDOS_WANNIER.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        data = np.genfromtxt(name)
    except:
        raise ParseError
    out = {'energy': data[:, 0], 'dos': data[:, 1]}

    return out


def parse_wannier_info(name: str) -> dict:
    """
    Parser for WANNIER_INFO.OUT

    :param str name: File name
    :return dict output: Parsed data
    """
    file = open(name)

    # Extract data
    lines = []
    data = []
    total = []
    start = False
    for i, line in enumerate(file.readlines()):
        if "* Wannier functions" in line:
            start = True
        if start:
            lines.append(line)
    for i in range(len(lines)):
        if lines[i].strip().startswith('1'):
            for j in range(4):
                data.append(lines[i + j].split())
        if lines[i].strip().startswith('5'):
            for j in range(4):
                data.append(lines[i + j].split())
        if lines[i].strip().startswith('total'):
            total.append(lines[i].split())
    file.close()

    # Package data into dictionary
    n_wannier = len(data)
    localisation_center = np.empty(shape=(n_wannier, 3))
    wannier = {
        'n_wannier': n_wannier,
        'Omega': [],
        'Omega_I': [],
        'Omega_D': [],
        'Omega_OD': []
    }

    for i, item in enumerate(data):
        localisation_center[i, :] = [float(x) for x in item[1:4]]
        wannier['Omega'].append(float(item[4]))
        wannier['Omega_I'].append(float(item[5]))
        wannier['Omega_D'].append(float(item[6]))
        wannier['Omega_OD'].append(float(item[7]))

    wannier['localisation_center'] = localisation_center

    totals = {'Omega': [], 'Omega_I': [], 'Omega_D': [], 'Omega_OD': []}
    for j, item in enumerate(total):
        totals['Omega'].append(float(item[1]))
        totals['Omega_I'].append(float(item[2]))
        totals['Omega_D'].append(float(item[3]))
        totals['Omega_OD'].append(float(item[4]))

    wannier['total'] = totals

    return wannier


def parse_core_overlap(name: str) -> dict:
    """
    Parser for coreoverlap.xml

    Parsed dictionary has the structure:

        output = {'nkpt':  nkpt
                  'nstfv': nstfv
                  'ncg':   ncg
                  'kpoints': [{'index': index, 'pairs': pairs},
                              {'index': index, 'pairs': pairs},
                              ...]
                  }

    where output['kpoints'][ik]['pairs'] =
      [{'ist1': '1', 'ist2': '1', 'de': 12.97849772, 'overlap': 3.35753859e-07},
       {'ist1': '1', 'ist2': '2', 'de': 12.97849772, 'overlap': 3.35753859e-07},
       ...
       n_pairs]

    :param str name: File name
    :return dict output: Parsed data
    """
    try:
        tree = ET.parse(name)
    except:
        raise ParseError

    root = tree.getroot()
    core_overlap = {
        'nkpt': int(root.attrib['nkpt']),
        'nstfv': int(root.attrib['nstfv']),
        'ncg': int(root.attrib['ncg'])
    }

    k_points = []
    for k_point in root:
        kpt = {"index": int(k_point.attrib['index'])}
        pairs = []
        for pair_xml in k_point:
            pair = pair_xml.attrib
            pair['ist1'] = int(pair['ist1'])
            pair['ist2'] = int(pair['ist2'])
            pair["de"] = float(pair["de"])
            pair["overlap"] = float(pair["overlap"].split()[0]) ** 2 + float(
                pair["overlap"].split()[1]) ** 2
            pairs.append(pair)
        kpt["pairs"] = pairs
        k_points.append(kpt)
    core_overlap["kpoints"] = k_points

    return core_overlap


def parse_lossfunction(fname: str) -> tuple:
    """
    Parses files containing loss function
    e.g. LOSS_FXCRPA_OC11_QMT001.OUT

    :param str fname: name of the file
    """
    xdata = []
    ydata = []
    file = open(fname, 'r')
    for lines in file:
        if 'Frequency' in lines:
            break
    for lines in file:
        data = lines.split()
        xdata.append(float(data[0]))
        ydata.append(float(data[1]))
    file.close()
    return xdata, ydata


def parse_wf1d(fname: str) -> dict:
  """
  Parses files containing one dimensional wave function plot as saved in
  the files, _e.g._, wf1d-0001-0001.dat, where the first 0001 indicates the k-point
  and the second the state index.
  
  :param str fname: name of the file
  """

  data = np.genfromtxt(fname)

  output = {}
  output["path"] = data[:, 0]
  output["|psi|^2"] = data[:, 1]
  output["Re(psi)"] = data[:, 2]
  output["Im(psi)"] = data[:, 3]

  return output 


def parse_wf2d(fname: str) -> dict:
  """
  Parses files containing the two dimensional wave function plot as saved in the files,
  _e.g._, wf2d-0001-0001.xsf, where the first 0001 indicates the k-point
  and the second the state index.

  Does not parse PRIMVEC and PRIMCOORD. These can be parsed with ase.io.read(fname).

  :param str fname: name of the file
  """

  file = open(fname, 'r')
  lines = file.readlines()
  output = {}
  
  i = 0
  while i < len(lines):
    if 'BEGIN_BLOCK_DATAGRID_2D' in lines[i]:
      i = i + 1
      data_name = lines[i].replace('\n', '').lstrip()
      
      i = i + 2
      grid = np.fromstring(lines[i], dtype=int, sep=' ')

      i = i + 1
      origin = np.fromstring(lines[i], dtype=np.double, sep = ' ')
      
      i = i + 1
      point1 = np.fromstring(lines[i], dtype=np.double, sep = ' ')
      
      i = i + 1
      point2 = np.fromstring(lines[i], dtype=np.double, sep = ' ')

      i = i + 1
      data = np.fromstring(lines[i], dtype=np.double, sep=' ')
      while 'END_DATAGRID_2D' not in lines[i]:
        i = i + 1
        new_data = np.fromstring(lines[i], dtype=np.double, sep=' ')
        data = np.concatenate((data, new_data), axis=None)
    
      output['grid'] = grid
      output['origin'] = origin
      output['point1'] = point1
      output['point2'] = point2
      output[data_name] = data # data_name: "module squared", "real", "imaginary"

    i = i + 1

  return output


def parse_wf3d(fname: str) -> dict:
  """
  Parses files containing the two dimensional wave function plot as saved in the files,
  _e.g._, wf3d-0001-0001.xsf, where the first 0001 indicates the k-point
  and the second the state index.

  Does not parse PRIMVEC and PRIMCOORD. These can be parsed with ase.io.read(fname).

  :param str fname: name of the file
  """

  file = open(fname, 'r')
  lines = file.readlines()
  output = {}
  
  i = 0
  while i < len(lines):
    if 'BEGIN_BLOCK_DATAGRID_3D' in lines[i]:
      i = i + 1
      data_name = lines[i].replace('\n', '').lstrip()
      
      i = i + 2
      grid = np.fromstring(lines[i], dtype=int, sep=' ')

      i = i + 1
      origin = np.fromstring(lines[i], dtype=np.double, sep = ' ')
      
      i = i + 1
      point1 = np.fromstring(lines[i], dtype=np.double, sep = ' ')
      
      i = i + 1
      point2 = np.fromstring(lines[i], dtype=np.double, sep = ' ')

      i = i + 1
      point3 = np.fromstring(lines[i], dtype=np.double, sep = ' ')

      i = i + 1
      data = np.fromstring(lines[i], dtype=np.double, sep=' ')
      while 'END_DATAGRID_3D' not in lines[i]:
        i = i + 1
        new_data = np.fromstring(lines[i], dtype=np.double, sep=' ')
        data = np.concatenate((data, new_data), axis=None)
    
      output['grid'] = grid
      output['origin'] = origin
      output['point1'] = point1
      output['point2'] = point2
      output['point3'] = point3
      output[data_name] = data # data_name: "module squared", "real", "imaginary"

    i = i + 1

  return output  

def parse_cube(fname: str) -> dict:
  """
  Parses .cube files. These files contain data calculated on a regular grid in a box.
  All vectors are given in cartesian coordinates. `output['cube_data']` contains the data. 
  It is described by `output['description']`.

  :param str fname: name of the file
  """

  file = open(fname, 'r')
  lines = file.readlines()
  output = {}

  output['title'] = str(lines[0])
  output['description'] = str(lines[1])

  n_atoms = int(lines[2].split()[0])
  output['n_atoms'] = n_atoms
  output['origin'] = np.array(lines[2].split()[1:], dtype=np.double)

  output['n_1'] = int(lines[3].split()[0])
  output['v_increment_1'] = np.array(lines[3].split()[1:], dtype=np.double)

  output['n_2'] = int(lines[4].split()[0])
  output['v_increment_2'] = np.array(lines[4].split()[1:], dtype=np.double)

  output['n_3'] = int(lines[5].split()[0])
  output['v_increment_3'] = np.array(lines[5].split()[1:], dtype=np.double)

  line_offset = 6
  output['atoms'] = []
  for line in lines[line_offset : line_offset + n_atoms]:
      atom_number = int(line.split()[0])
      charge = np.double(line.split()[1])
      coordinate = np.array(line.split()[2:], dtype=np.double)
      output['atoms'].append(dict(atom_number=atom_number, charge=charge, coordinate=coordinate))

  line_offset = line_offset + n_atoms

  cube_data = []
  for line in lines[line_offset:]:
      cube_data = cube_data + line.split()
  
  output['cube_data'] = np.array(cube_data, dtype=np.double)

  return output
  



""" Main exciting input classes. """
from excitingtools.input.input_classes import ExcitingGroundStateInput, ExcitingXSInput, ExcitingInputXML, \
    ExcitingPropertiesInput
from excitingtools.input.structure import ExcitingStructure



"""
Parsers for real-time TDDFT output files
"""
from xml.etree.ElementTree import ParseError
import numpy as np
from typing import List

from excitingtools.parser_utils.grep_parser import grep


def parse_nexc(name, skiprows=1):
    """
    Parser for NEXC.OUT
    """
    try:
        data = np.genfromtxt(name, skip_header=skiprows)
    except:
        raise ParseError
    out = {
        "Time": data[:, 0],
        "number_electrons_GroundState": data[:, 1],
        "number_electrons_ExcitedState": data[:, 2],
        "sum": data[:, 3]
    }

    return out


def parse_jind(name, skiprows=0):
    """
    Parser for JIND.OUT
    """
    try:
        data = np.genfromtxt(name, skip_header=skiprows)
    except:
        raise ParseError
    out = {
        "Time": data[:, 0],
        "Jx": data[:, 1],
        "Jy": data[:, 2],
        "Jz": data[:, 3]
    }

    return out


def parse_etot(name):
    """
    Parser for ETOT_RTTDDFT.OUT
    """
    try:
        data = np.genfromtxt(name, skip_header=1)
    except:
        raise ParseError
    out = {
        "Time": data[:, 0],
        "ETOT": data[:, 1],
        "Madelung": data[:, 2],
        "Eigenvalues-Core": data[:, 3],
        "Eigenvalues-Valence": data[:, 4],
        "Exchange": data[:, 5],
        "Correlation": data[:, 6],
        "XC-potential": data[:, 7],
        "Coulomb pot. energy": data[:, 8]
    }

    return out


def parse_eigval_screenshots(name: str) -> dict:
    """
    Parser for EIGVAL_*.OUT.
    """

    def get_k_point_blocks(name: str, fortran_index=False) -> List[dict]:
        """
        Parse the k point blocks.

        For a file:
        line 0   ik =       1
        line 1    1     -0.357413438539
                  .
        line 35   35      8.032537750481
        line 36
        line 37  ik =       2
        line 38   1     -0.280187464596
                  .
        line 75   38      8.808472531393

        The parsed k-point block lines correspond to:
          k_blocks[0]['start'] = line 0   i.e.   ik =       1
          k_blocks[0]['end']   = line 35  i.e.   35      8.032537750481
          k_blocks[1]['start'] = line 37  i.e.   ik =       2
          k_blocks[1]['end']   = line 75  i.e.   38      8.808472531393


        Default line indexing starts at 0.
        """

        # Lines for which a new k-point block starts
        raw_k_point_lines = grep("ik", name, options={'n': ''}).splitlines()
        k_point_lines = [int(line.split(':')[0]) for line in raw_k_point_lines]

        if fortran_index:
            offset = 1
        else:
            # Python indexing
            offset = 0
            k_point_lines = [int(i) - 1 for i in k_point_lines]

        n_lines = sum(1 for line in open(name)) + offset

        k_blocks = []
        k_start = 0 + offset
        for ik in k_point_lines[1:]:
            k_end = ik - 2
            k_blocks.append({'start': k_start, 'end': k_end})
            k_start = k_end + 2

        # Account for final k-block
        k_blocks.append({'start': k_start, 'end': n_lines - 2})

        return k_blocks

    k_blocks = get_k_point_blocks(name)

    # Parse file
    with open(name, 'r') as f:
        file = f.readlines()

    data = {}
    kpoints = []
    for indices in k_blocks:
        kpoint = {}
        ik = int(file[indices['start']].split()[-1])
        eigenvalues = [
            float(file[i].split()[-1])
            for i in range(indices['start'] + 1, indices['end'] + 1)
        ]
        kpoint['ik'] = ik
        kpoint['eigenvalues'] = eigenvalues
        kpoints.append(kpoint)

    data['kpoints'] = kpoints

    return data


def parse_proj_screenshots(name: str) -> dict:
    """
    Parser for PROJ_*.OUT.

    Effectively the same code as parse_eigval_screenshots, but the whitespace
    between blocks differs by 1.
    """

    raw_k_point_lines = grep("ik", name, options={'n': ''}).splitlines()
    k_point_lines = [int(line.split(':')[0]) - 1 for line in raw_k_point_lines]
    last_line = sum(1 for line in open(name))

    k_blocks = []
    k_start = 0
    for ik in k_point_lines[1:]:
        k_end = ik
        k_blocks.append([k_start, k_end])
        k_start = k_end
    k_blocks.append([k_end, last_line])

    # Parse file
    with open(name) as f:
        file = f.readlines()

    data = {}
    kpoints = []
    for i in k_blocks:
        kpoint = {}
        start = i[0]
        end = i[1]
        ik = int(file[start].split()[-1])
        projection = []
        for j in range(start + 1, end):
            projection.append([float(x) for x in file[j].split()])
        kpoint['ik'] = ik
        kpoint['projection'] = np.asarray(projection)

    data['kpoints'] = kpoints

    return data

def parse_atom_position_velocity_force(name: str) -> dict:
    """ Parser for ATOM_????.OUT 
    :param str name: name of file to parse
    :return dict out: each dict key corresponds to time, position (3 columns), velocity (3 columns), total force (3 columns).
    The 3 columns refer to the x, y, z components
    """
    try:
        data = np.genfromtxt(name, skip_header=0)
    except:
        raise ParseError
    out = {
        "Time": data[:, 0],
        "x": data[:, 1],
        "y": data[:, 2],
        "z": data[:, 3],
        "vx": data[:, 4],
        "vy": data[:, 5],
        "vz": data[:, 6],
        "Fx": data[:, 7],
        "Fy": data[:, 8],
        "Fz": data[:, 9]
    }

    return out

def parse_force(name, skiprows=0):
    """
    Parser for X_????.OUT, where X can be:
    - FCR: core corrections to forces
    - FEXT: external forces (due to e.g. an electric field)
    - FHF: Hellman-Feynman term of forces
    - FVAL: valence corrections to forces
    """
    try:
        data = np.genfromtxt(name, skip_header=skiprows)
    except:
        raise ParseError
    out = {
        "Time": data[:, 0],
        "Fx": data[:, 1],
        "Fy": data[:, 2],
        "Fz": data[:, 3]
    }

    return out


"""Test ExcitingGroundStateInput class attribute assignment and methods.

NOTE:
All attribute tests should assert on the XML tree content's as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.
"""
import numpy as np
import pytest

# noinspection PyUnresolvedReferences
from excitingtools.input.input_classes import ExcitingBSEInput, ExcitingXSInput


def test_class_ExcitingXSInput():
    xs_input = ExcitingXSInput(xstype="BSE")

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'
    assert xs_xml.items() == [('xstype', 'BSE')]


def test_class_ExcitingXSInput_xstype_missing():
    with pytest.raises(ValueError, match="Missing mandatory arguments: {'xstype'}"):
        ExcitingXSInput(ngridk=[4, 4, 4])


def test_ExcitingXSInput_as_dict(mock_env_jobflow_missing):
    xs_input = ExcitingXSInput(xstype="BSE")
    ref_dict = {'xml_string': '<xs xstype="BSE"> </xs>'}
    assert xs_input.as_dict() == ref_dict, 'expected different dict representation'


def test_ExcitingXSInput_as_dict_jobflow(mock_env_jobflow):
    xs_input = ExcitingXSInput(xstype="BSE")
    ref_dict = {'@class': 'ExcitingXSInput',
                '@module': 'excitingtools.input.input_classes',
                'xml_string': '<xs xstype="BSE"> </xs>'}
    assert xs_input.as_dict() == ref_dict, 'expected different dict representation'


def test_ExcitingXSInput_from_dict():
    ref_dict = {'xml_string': '<xs xstype="BSE"> </xs>'}
    recreated_xs = ExcitingXSInput.from_dict(ref_dict)
    assert recreated_xs.to_xml_str() == ref_dict['xml_string']


def test_class_ExcitingXSInput_xs():
    xs = {'broad': 0.32, 'ngridk': [8, 8, 8], 'tevout': True, 'nempty': 52, 'pwmat': 'fft', "xstype": "BSE"}
    xs_input = ExcitingXSInput(**xs)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'
    assert set(xs_xml.keys()) == set(xs)
    assert xs_xml.get('xstype') == 'BSE'
    assert xs_xml.get('broad') == '0.32'
    assert xs_xml.get('ngridk') == '8 8 8'
    assert xs_xml.get('tevout') == 'true'
    assert xs_xml.get('nempty') == '52'
    assert xs_xml.get('pwmat') == 'fft'


def test_class_ExcitingXsInput_wrong_key():
    with pytest.raises(ValueError, match="xs keys are not valid: {'wrong_key'}"):
        ExcitingXSInput(xstype="BSE", wrong_key=1)


def test_class_ExcitingXSInput_BSE_element():
    bse_attributes = {'bsetype': 'singlet', 'xas': True, 'xasspecies': 1}
    bse_keys = list(bse_attributes)
    xs_input = ExcitingXSInput(xstype="BSE", BSE=bse_attributes)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'
    assert set(xs_xml.keys()) == {'xstype'}

    elements = list(xs_xml)
    assert len(elements) == 1

    bse_xml = elements[0]
    assert bse_xml.tag == "BSE"
    assert bse_xml.keys() == bse_keys, 'Should contain all bse attributes'
    assert bse_xml.get('bsetype') == 'singlet'
    assert bse_xml.get('xas') == 'true'
    assert bse_xml.get('xasspecies') == '1'


def test_class_ExcitingXSInput_BSE_element_object():
    bse_object = ExcitingBSEInput(bsetype='singlet', xas=True, xasspecies=1)
    bse_keys = {'bsetype', 'xas', 'xasspecies'}
    xs_input = ExcitingXSInput(xstype="BSE", BSE=bse_object)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'
    assert set(xs_xml.keys()) == {'xstype'}

    elements = list(xs_xml)
    assert len(elements) == 1

    bse_xml = elements[0]
    assert bse_xml.tag == "BSE"
    assert set(bse_xml.keys()) == bse_keys, 'Should contain all bse attributes'
    assert bse_xml.get('bsetype') == 'singlet'
    assert bse_xml.get('xas') == 'true'
    assert bse_xml.get('xasspecies') == '1'


def test_class_ExcitingXSInput_energywindow_element():
    energywindow_attributes = {'intv': [5.8, 8.3], 'points': 5000}
    xs_input = ExcitingXSInput(xstype="BSE", energywindow=energywindow_attributes)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'

    elements = list(xs_xml)
    assert len(elements) == 1

    energywindow_xml = elements[0]
    assert energywindow_xml.tag == "energywindow"
    assert set(energywindow_xml.keys()) == set(energywindow_attributes), 'Should contain all bse attributes'
    assert energywindow_xml.get('intv') == '5.8 8.3'
    assert energywindow_xml.get('points') == '5000'


def test_class_ExcitingXSInput_screening_element():
    screening_attributes = {'screentype': 'full', 'nempty': 15}
    xs_input = ExcitingXSInput(xstype="BSE", screening=screening_attributes)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'

    elements = list(xs_xml)
    assert len(elements) == 1

    screening_xml = elements[0]
    assert screening_xml.tag == "screening"
    assert set(screening_xml.keys()) == set(screening_attributes), 'Should contain all bse attributes'
    assert screening_xml.get('screentype') == 'full'
    assert screening_xml.get('nempty') == '15'


def test_class_ExcitingQpointsetInput_numpy():
    qpointset_input = np.array(((0, 0, 0), (0.5, 0.5, 0.5)))

    xs_input = ExcitingXSInput(xstype="BSE", qpointset=qpointset_input)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'

    elements = list(xs_xml)
    assert len(elements) == 1

    qpointset_xml = elements[0]
    assert qpointset_xml.tag == "qpointset"
    assert qpointset_xml.items() == []

    qpoints = list(qpointset_xml)
    assert len(qpoints) == 2
    assert qpoints[0].tag == 'qpoint'
    assert qpoints[0].items() == []
    assert qpoints[0].text == '0.0 0.0 0.0'
    assert qpoints[1].text == '0.5 0.5 0.5'


def test_class_ExcitingQpointsetInput_list():
    qpointset_input = [[0, 0, 0], [0.5, 0.5, 0.5]]

    xs_input = ExcitingXSInput(xstype="BSE", qpointset=qpointset_input)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'

    elements = list(xs_xml)
    assert len(elements) == 1

    qpointset_xml = elements[0]
    assert qpointset_xml.tag == "qpointset"
    assert qpointset_xml.items() == []

    qpoints = list(qpointset_xml)
    assert len(qpoints) == 2
    assert qpoints[0].tag == 'qpoint'
    assert qpoints[0].items() == []
    assert qpoints[0].text == '0 0 0'
    assert qpoints[1].text == '0.5 0.5 0.5'


def test_class_ExcitingPlanInput():
    plan_input = ['screen', 'bse', 'bsegenspec']

    xs_input = ExcitingXSInput(xstype="BSE", plan=plan_input)

    xs_xml = xs_input.to_xml()
    assert xs_xml.tag == 'xs'

    elements = list(xs_xml)
    assert len(elements) == 1

    plan_xml = elements[0]
    assert plan_xml.tag == "plan"
    assert plan_xml.items() == []

    doonlys = list(plan_xml)
    assert len(doonlys) == 3
    assert doonlys[0].tag == 'doonly'
    assert doonlys[0].items() == [('task', 'screen')]
    assert doonlys[1].tag == 'doonly'
    assert doonlys[1].items() == [('task', 'bse')]
    assert doonlys[2].tag == 'doonly'
    assert doonlys[2].items() == [('task', 'bsegenspec')]


def test_class_ExcitingPlanInput_wrong_plan():
    plan_input = ['screen', 'bse', 'bsegenspec', 'falseplan']

    with pytest.raises(ValueError) as error:
        ExcitingXSInput(xstype="BSE", plan=plan_input)
    assert error.value.args[0] == "plan keys are not valid: {'falseplan'}"


def test_class_ExcitingXSInput_attribute_setting_getting():
    xs_input = ExcitingXSInput(xstype="BSE")

    xs_input.ngridk = [2, 2, 2]
    with pytest.raises(ValueError, match="xs keys are not valid: {'abc'}"):
        xs_input.abc = 3


def test_class_ExcitingXSInput_attribute_deleting():
    xs_input = ExcitingXSInput(xstype="BSE", ngridk=[2, 2, 2])

    del xs_input.ngridk
    with pytest.warns(UserWarning, match="Attempt to delete mandatory attribute 'xstype' was prevented."):
        del xs_input.xstype  # pylint: disable=no-member



"""Utils for serialization for jobflow.
"""
import os


def special_serialization_attrs(instance) -> dict:
    """Gives special keys + values for serialization.
     Currently only supports jobflow via the env_var USE_JOBFLOW.

    :param instance: object you want to serialise
    :returns: dictionary with the special keys + values
    """
    if os.getenv('USE_JOBFLOW') is not None:
        return {"@module": instance.__class__.__module__, "@class": instance.__class__.__name__, }
    return {}



""" Classes and functions to aid unit testing.
"""
import pathlib


class MockFile:
    """ Single class for testing parsers that require either:
       * File.
       * String contents of file.

    Usage:
    @pytest.fixture
    def file_mock(tmp_path):
       file = tmp_path / _file_name
       file.write_text(string_contents)
       return MockFile(file, string_contents)
    """
    def __init__(self, file: pathlib.Path, string: str):
        # File object
        self.file = file
        # File contents
        self.string = string
        # Name prepended by path
        self.full_path = self.file.as_posix()



""" Parse exciting species files into dictionary.
"""
from typing import Dict

from excitingtools.parser_utils.parser_decorators import xml_root
from excitingtools.utils.dict_utils import string_value_to_type
from excitingtools.utils.utils import string_to_bool


@xml_root
def parse_species_xml(root) -> dict:
    """ Parses exciting species files as a dict.

    TODO(Alex) Issue 124. See how easy it is to replace with a generic XML
    parser, with keys defined according to the associated schema.

    Return a dictionary with elements:

      species = {'chemicalSymbol': chemicalSymbol, 'name': name, 'z': z, 'mass': mass}

      muffin_tin = {'rmin': rmin, 'rinf': rinf, 'radius': radius, 'points':  radialmeshPoints}

      atomic_states = [{'n': 1, 'l': 0, 'kappa': 1, 'occ': 2.0, 'core': True},
                      {'n': 2, 'l': 0, 'kappa': 1, 'occ': 2.0, 'core': True}, ...]

      basis['default'] = [{'type': 'lapw', 'trialEnergy': '0.1500', 'searchE': 'true'}]

      basis['custom'] = [{'l': 0, 'type': 'lapw', 'trialEnergy': 1.35670550183736, 'searchE': False},
                         {'l': 1, 'type': 'lapw', 'trialEnergy': -2.69952312512447, 'searchE': False},
                         {'l': 2, 'type': 'lapw', 'trialEnergy': 0.00,  'searchE': False},
                         {'l': 3, 'type': 'lapw', 'trialEnergy': 1.000, 'searchE': False},
                         {'l': 4, 'type': 'lapw', 'trialEnergy': 1.000, 'searchE': False},
                         {'l': 5, 'type': 'lapw', 'trialEnergy': 1.000, 'searchE': False}]

      basis['lo'] = [{'l': 0, 'matchingOrder': [0, 1], 'trialEnergy': [-4.3784, -4.3784], 'searchE': [False, False]},
                     {'l': 0, 'matchingOrder': [0, 1], 'trialEnergy': [1.356, 1.3566], 'searchE': [False, False]},
                    ...]

    :param root: XML file, XML string, or an ET.Element.
    :return : Dictionary of species file data (described above).
    """
    species_tree = root[0]
    species = {key: value for key, value in species_tree.attrib.items()}

    for key in ['z', 'mass']:
        species[key] = float(species[key])

    children: Dict[str, list] = {'atomicState': [], 'basis': [], 'muffinTin': []}
    for child in list(species_tree):
        children[child.tag].append(child)

    assert len(children['muffinTin']) == 1, "More than one muffinTin sub-tree in the species file"
    assert len(children['basis']) == 1, "More than one basis sub-tree in the species file"

    muffin_tin_tree = children['muffinTin'][0].attrib
    muffin_tin = {key: float(value) for key, value in muffin_tin_tree.items()}

    atomic_states = []
    for atomic_state_tree in children['atomicState']:
        assert atomic_state_tree.tag == 'atomicState', "Expect tag to be atomicState"
        atomic_states.append(string_value_to_type(atomic_state_tree.attrib))

    basis_tree = children['basis'][0]
    basis: Dict[str, list] = {'default': [], 'custom': [], 'lo': []}

    for func in basis_tree:
        function: dict = func.attrib
        processed_function = string_value_to_type(function)

        if func.tag == 'lo':
            processed_function.update(_parse_lo_from_species(func))

        basis[func.tag].append(processed_function)

    return {
        'species': species,
        'muffin_tin': muffin_tin,
        'atomic_states': atomic_states,
        'basis': basis
        }


def _parse_lo_from_species(lo_function) -> dict:
    """
    Given some lo_function with:
      wf {'matchingOrder': '0', 'trialEnergy': '-2.0', 'searchE': 'true'}
      wf {'matchingOrder': '1', 'trialEnergy': '-2.0', 'searchE': 'true'}

    return
    {'matchingOrder': [0, 1], 'trialEnergy': [-2.0, -2.0], 'searchE': [True, True]}
    """
    # Use lists to GUARANTEE consistent ordering
    matching_order = []
    trial_energy = []
    search = []
    for radial in lo_function:
        matching_order.append(int(radial.attrib.get('matchingOrder')))
        trial_energy.append(float(radial.attrib.get('trialEnergy')))
        search.append(string_to_bool(radial.attrib.get('searchE')))
    return {'matchingOrder': matching_order, 'trialEnergy': trial_energy, 'searchE': search}



""" GW eigenvalue parser, processing dict values and returning to an object.
"""
import os
from typing import Optional, List, Union, Dict
import enum
import numpy as np

from excitingtools.exciting_dict_parsers.gw_eigenvalues_parser import parse_evalqp, _file_name, parse_gw_dos
from excitingtools.dataclasses.data_structs import NumberOfStates
from excitingtools.dataclasses.eigenvalues import EigenValues
from excitingtools.dataclasses.density_of_states import DOS


class NitrogenEvalQPColumns(enum.Enum):
    E_KS = 0
    E_HF = 1
    E_GW = 2
    Sx = 3
    Sc = 4
    Vxc = 5
    DE_HF = 6
    DE_GW = 7
    Znk = 8


class OxygenEvalQPColumns(enum.Enum):
    """Columns of `_file_name`, for exciting oxygen
    excluding the state index."""
    E_KS = 0
    E_HF = 1
    E_GW = 2
    Sx = 3
    Re_Sc = 4
    Im_Sc = 5
    Vxc = 6
    DE_HF = 7
    DE_GW = 8
    Znk = 9


# Columns pass as a single enum, or list of enums.
columns_type = Union[enum.Enum, List[enum.Enum]]

# Return an instance of EigenValues, or a dicts of EigenValues if the user asks for
# multiple data columns.
return_type = Union[EigenValues, Dict[enum.Enum, EigenValues]]


def gw_eigenvalue_parser(input_file_path: str, columns: Optional[columns_type] = OxygenEvalQPColumns.E_GW) -> \
        return_type:
    """ High-level Parser for GW eigenvalues file.

    Unpacks the result of dict into a sensible form and returns the data to return_type.

    :param input_file_path: File path (can include or exclude file name).
    :param columns: Optional choice for which data column of energies to return. Default is to return
    GW eigenvalues, assuming exciting oxygen (most recent release).
    :return An instance of EigenValues, or a dict of (key, value) = [EvalQPColumns, EigenValues].
    """
    path, file_name = os.path.split(input_file_path)
    if file_name == _file_name:
        file_path = path
    else:
        file_path = input_file_path
    abs_file_name = os.path.join(file_path, _file_name)

    if not isinstance(columns, list):
        columns = [columns]

    # Parse data
    data: dict = parse_evalqp(abs_file_name)
    state_range_list: List[int] = data.pop('state_range')
    column_enums = data.pop('column_labels')

    n_k = len(data.keys())
    state_range = NumberOfStates(state_range_list[0], state_range_list[1])

    # Check for inconsistency between requested column and columns available (i.e. data produced with
    # different code versions)
    # Because the parsed column data class is dynamically-generated, one has to do it with lengths
    parsed_column_names = {enum.value for enum in column_enums}
    requested_column_names = {enum.value for enum in type(columns[0])}

    if (requested_column_names - parsed_column_names) != set():
        enum_class_name = type(columns[0]).__name__
        raise ValueError(f'The requested data column is indexed according to exciting version {enum_class_name},'
                         f'which is not consistent with the columns of the parsed data.'
                         f' Check that your data was produced with the same code version.')

    # Repackage data
    k_indices = []
    k_points = []
    weights = []
    n_columns = len(type(columns[0]))
    all_eigenvalues = np.empty(shape=(n_k, state_range.n_states, n_columns))

    for ik, k_block in data.items():
        k_indices.append(ik)
        k_points.append(k_block['k_point'])
        weights.append(k_block['weight'])
        all_eigenvalues[ik - 1, :, :] = k_block['energies'][:, :]

    # Return data
    if len(columns) == 1:
        return EigenValues(state_range, k_points, k_indices, all_eigenvalues[:, :, columns[0].value], weights)

    eigen_values = {}
    for column in columns:
        # Note, bad memory access pattern for all_eigenvalues, but would need to transpose in parse_evalqp, then
        # have bad access in the ik loop above
        value = EigenValues(state_range, k_points, k_indices, all_eigenvalues[:, :, column.value], weights)
        eigen_values[column] = value

    return eigen_values


def parse_obj_gw_dos(full_file_name: str) -> DOS:
    """High-level parser for GW DOS files.

    :param full_file_name: Path + file name
    :return: DOS object
    """
    gw_dos_data = parse_gw_dos(full_file_name)
    return DOS(gw_dos_data['energy'], gw_dos_data['dos'])



"""
Test regex parser wrappers
"""
import pytest

from excitingtools.parser_utils.regex_parser import parse_value_regex


def test_parse_value_regex():

    test_string = """
    --------------------------------------------------------------------------------
    -                               frequency grid                                 -
    --------------------------------------------------------------------------------
     
     Type: < fgrid > gauleg2                                 
     Frequency axis: < fconv > imfreq                                  
     Number of frequencies: < nomeg >           32
     Cutoff frequency: < freqmax >    1.00000000000000  
    
    --------------------------------------------------------------------------------
    -                          Kohn-Sham band structure                            -
    --------------------------------------------------------------------------------
     
     Fermi energy:     0.0000
     Energy range:   -14.6863 1030.7919
     Band index of VBM:  21
     Band index of CBm:  22
     
     Indirect BandGap (eV):                    3.3206
     at k(VBM) =    0.000   0.500   0.500 ik =     3
        k(CBm) =    0.000   0.000   0.000 ik =     1
     Direct Bandgap at k(VBM) (eV):            3.7482
     Direct Bandgap at k(CBm) (eV):            3.8653
     
     --------------------------------------------------------------------------------
     -                            G0W0 band structure                               -
     --------------------------------------------------------------------------------
      
      Fermi energy:    -0.0054
      Energy range:   -16.2632 1031.4090
      Band index of VBM:  21
      Band index of CBm:  22
      
      Indirect BandGap (eV):                    5.3920
      at k(VBM) =    0.000   0.500   0.500 ik =     3
         k(CBm) =    0.000   0.000   0.000 ik =     1
      Direct Bandgap at k(VBM) (eV):            5.5472
      Direct Bandgap at k(CBm) (eV):            5.9646
    """

    assert {'Fermi energy': 0.0} == parse_value_regex(test_string, 'Fermi energy:'), \
        "Expect to match first string instance and return {key:float}"

    assert {'Indirect BandGap (eV)': 3.3206} == parse_value_regex(test_string, 'Indirect BandGap \\(eV\\):'), \
       "Expect to match first string instance and return {key:float}"

    assert {'Energy range': [-14.6863, 1030.7919]} == parse_value_regex(test_string, 'Energy range:'), \
        "Expect to match first string instance and return {key:List[float]}"

    assert {'Type: < fgrid >': 'gauleg2'} == parse_value_regex(test_string, 'Type: < fgrid >'), \
        "Expect to match first string instance and return {key:str}"


@pytest.mark.xfail
def test_failures_with_parse_value_regex():
    """
    This fails for timings because matching 'Initialization' returns ':        15.46',
    (for example), and the colon cannot be interpreted by eval()
    """
    test_string = """
     Initialization                             :        15.46
         - init_scf                             :         8.38
    """

    data = parse_value_regex(test_string, 'Initialization')



"""General parser utility functions.
"""
import re
from typing import Dict, Any
from xml.etree import ElementTree
from json import JSONDecodeError, loads


def find_element(root: ElementTree.Element, tag: str) -> ElementTree.Element:
    """ Finds a given tag in an Element, either return the full ElementTree
    if the tag is correct or find the tag in that ElementTree.
    :param root: Element to find the tag in
    :param tag: tag to search for
    :return: the (sub)element with the correct tag
    """
    if root.tag == tag:
        return root
    return root.find(tag)


def convert_string_dict(inputs: dict) -> Dict[str, Any]:
    """ Parses and converts a dictionary with string values to their actual data types.

    :param inputs: input dictionary
    :return: the converted dictionary
    """
    for key, value in inputs.items():
        inputs[key] = convert_single_entry(value)
    return inputs


def convert_single_entry(input_str: str):
    """ Converts a single string. Accepts also lists separated by whitespaces.

    :param input_str: input string
    :return: the converted data
    """
    split_string = input_str.split()
    result = []
    for i in split_string:
        result.append(json_convert(standardise_fortran_exponent(i)))

    if len(result) == 1:
        return result[0]
    if len(list((filter(lambda x: isinstance(x, str), result)))) == len(result):
        return " ".join(result)
    return result


def json_convert(input_str: str):
    """ Tries to convert a single string with no whitespaces to its actual data type
    using the json decoder (detects int, float and bool). Else returns the string as string.
    :param input_str: input string
    :return: the converted value
    """
    try:
        return loads(input_str)
    except JSONDecodeError:
        return input_str


def standardise_fortran_exponent(input_str: str, return_as_str: bool = True):
    """ Tries to convert a single string representing a float value in scientific notation
    to the actual number.
    d(D) and q(Q) correspond to higher floating point precision than e(E). Replace
    them since they cannot be parsed by JSON

    :param input_str: input string
    :param return_as_str: return the number as a string in python format with e as exponential
    :return: If string can be converted to a float return it, else
    return the input string.
    """
    subbed_str = re.sub('[dDqQ]', 'E', input_str, count=1)
    if subbed_str == input_str:
        return input_str

    try:
        number = float(subbed_str)
    except ValueError:
        return input_str
    if return_as_str:
        return str(number)
    return number



# Units
from excitingtools.constants.units import Unit
# Parsers returning to dicts
# Questionable whether one should expose this - required for test framework recursive comparisons
# Typically not the API one wants to expose to the user, as parsed dict keys are subject to change
from excitingtools.exciting_dict_parsers.parser_factory import parser_chooser
# Parsers returning to objects
from excitingtools.exciting_obj_parsers import *
# User-level objects
from excitingtools.dataclasses import *
# Input objects
from excitingtools.input import *

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

__version__ = metadata.version("excitingtools")
